"""fault.py: Module providing resqml classes relating to fault representation."""

version = '2nd July 2021'

# Nexus is a registered trademark of the Halliburton Company

import logging
log = logging.getLogger(__name__)
log.debug('fault.py version ' + version)

import numpy as np
import pandas as pd
import os
# import xml.etree.ElementTree as et
# from lxml import etree as et

from resqpy.olio.base import BaseResqpy
import resqpy.olio.read_nexus_fault as rnf
import resqpy.olio.xml_et as rqet
import resqpy.olio.write_hdf5 as rwh5
import resqpy.olio.uuid as bu
import resqpy.olio.trademark as tm
from resqpy.olio.xml_namespaces import curly_namespace as ns

import resqpy.organize as rqo



class GridConnectionSet(BaseResqpy):
   """Class for obj_GridConnectionSetRepresentation holding pairs of connected faces, usually for faults."""

   resqml_type = 'GridConnectionSetRepresentation'

   def __init__(self, parent_model, uuid = None, connection_set_root = None, grid = None,
                ascii_load_format = None, ascii_file = None,
                k_faces = None, j_faces = None, i_faces = None, feature_name = None,
                create_organizing_objects_where_needed = False,
                title = None, originator = None, extra_metadata = None):
      """Initializes a new GridConnectionSet and optionally loads it from xml or a list of simulator format ascii files.

      arguments:
         parent_model (model.Model object): the resqml model that this grid connection set will be part of
         uuid (uuid.UUID, optional): the uuid of an existing RESQML GridConnectionSetRepresentation from which
               this resqpy object is populated
         connection_set_root (DEPRECATED): use uuid instead; the root node of the xml tree for the
               obj_GridConnectionSet part; ignored if uuid is present
         grid (grid.Grid object, optional): If present, the grid object that this connection set relates to;
               if absent, the main grid for the parent model is assumed; only used if connection set root is
               None; see also notes
         ascii_load_format (string, optional): If present, must be 'nexus'; ignored if loading from xml;
               otherwise required if ascii_file is present
         ascii_file (string, optional): the full path of an ascii file holding fault definition data in
               nexus keyword format; ignored if loading from xml; otherwise, if present, ascii_load_format
               must also be set
         k_faces, j_faces, i_faces (boolean arrays, optional): if present, these arrays are used to identify
               which faces between logically neighbouring cells to include in the new grid connection set
         create_organizing_objects_where_needed (boolean, default False): if True when loading from ascii or
               face masks, a fault interpretation object and tectonic boundary feature object will be created
               for any named fault for which such objects do not exist; if False, missing organizational objects
               will cause an error to be logged; ignored when loading from xml
         title (str, optional): the citation title to use for a new grid connection set;
            ignored if uuid or connection_set_root is not None
         originator (str, optional): the name of the person creating the new grid connection set, defaults to login id;
            ignored if uuid or connection_set_root is not None
         extra_metadata (dict, optional): string key, value pairs to add as extra metadata for the grid connection set
            ignored if uuid or connection_set_root is not None

      returns:
         a new GridConnectionSet object, initialised from xml or ascii file, or left empty

      notes:
         in the resqml standard, a grid connection set can be used to identify connections between different
         grids (eg. parent grid to LGR); however, many of the methods in this code currently only handle
         connections within a single grid;
         when loading from an ascii file, cell faces are paired on a simplistic logical neighbour basis (as if
         there is no throw on the fault); this is because the simulator input does not include the full
         juxtaposition information; the simple mode is adequate for identifying which faces are involved in a
         fault but not for matters of juxtaposition or absolute transmissibility calculations;
         if uuid is None and connection_set_root is None and ascii_file is None and k_faces, j_faces & i_faces are None,
         then an empty connection set is returned

      :meta common:
      """

      log.debug('initialising grid connection set')
      self.count = None              # number of face-juxtaposition pairs in this connection set
      self.cell_index_pairs = None   # shape (count, 2); dtype int; index normalized for flattened array
      self.cell_index_pairs_null_value = -1  # integer null value for array above
      self.grid_index_pairs = None   # shape (count, 2); dtype int; optional; used if more than one grid referenced
      self.face_index_pairs = None   # shape (count, 2); dtype int32; local to cell, ie. range 0 to 5
      self.face_index_pairs_null_value = -1  # integer null value for array above
      # NB face index values 0..5 usually mean [K-, K+, J+, I+, J-, I-] respectively but there is some ambiguity
      #    over I & J in the Energistics RESQML Usage Guide; see comments in DevOps backlog item 269001 for more info
      self.grid_list = []            # ordered list of grid objects, indexed by grid_index_pairs
      self.feature_indices = None    # shape (count,); dtype int; optional; which fault interpretation each pair is part of
      # note: resqml data structures allow a face pair to be part of more than one fault but this code restricts to one
      self.feature_list = None       # ordered list, actually of interpretations, indexed by feature_indices
      # feature list contains tuples: (content_type, uuid, title) for fault features (or other interpretations)

      # NB: RESQML documentation is not clear which order is correct; should be kept consistent with same data in property.py
      # face_index_map maps from (axis, p01) to face index value in range 0..5
      # this is the default as indicated on page 139 (but not p. 180) of the RESQML Usage Gude v2.0.1
      # also assumes K is generally increasing downwards
      # see DevOps backlog item 269001 discussion for more information
#     self.face_index_map = np.array([[0, 1], [4, 2], [5, 3]], dtype = int)
      self.face_index_map = np.array([[0, 1], [2, 4], [5, 3]], dtype = int)  # order: top, base, J-, I+, J+, I-
      # and the inverse, maps from 0..5 to (axis, p01)
#     self.face_index_inverse_map = np.array([[0, 0], [0, 1], [1, 1], [2, 1], [1, 0], [2, 0]], dtype = int)
      self.face_index_inverse_map = np.array([[0, 0], [0, 1], [1, 0], [2, 1], [1, 1], [2, 0]], dtype = int)
      # note: the rework_face_pairs() method, below, overwrites the face indices based on I, J cell indices
      if not title: title = feature_name

      super().__init__(model = parent_model, uuid = uuid, title = title, originator = originator,
                       extra_metadata = extra_metadata, root_node = connection_set_root)

      if self.root is None:
         log.debug('setting grid for new connection set to default (ROOT)')
         if grid is None:
            grid = self.model.grid(find_properties = False)  # should find only or ROOT grid
            assert grid is not None, 'No ROOT grid found in model'
         self.grid_list = [grid]
         if ascii_load_format and ascii_file:
            if ascii_load_format == 'nexus':
               log.debug('loading connection set (fault) faces from Nexus format ascii file: ' + ascii_file)
               tm.log_nexus_tm('debug')
               faces = rnf.load_nexus_fault_mult_table(ascii_file)
            else:
               log.debug('ascii format for connection set faces not handled by base resqpy code: ' + ascii_load_format)
            assert faces is not None, 'failed to load fault face information from file: ' + ascii_file
            self.set_pairs_from_faces_df(faces, create_organizing_objects_where_needed)
            # note: hdf5 write and xml generation elsewhere
            # todo: optionally set face sets in grid object? or add to existing set?
            # grid.make_face_set_from_dataframe(faces) to set from dataframe
            # to access pair of lists for j & i plus faces, from grid attibutes, eg:
            # for fs in grid.face_set_dict.keys():
            #    list_pair = grid.face_set_dict[fs][0:2]
            #    face_sets.append(str(fs) + ': ' + str(len(list_pair[1])) + ' i face cols; ' +
            #                                      str(len(list_pair[0])) + ' j face cols; from ' + origin)
            # note: grid structures assume curtain like set of kelp
         elif k_faces is not None or j_faces is not None or i_faces is not None:
            self.set_pairs_from_face_masks(k_faces, j_faces, i_faces, feature_name, create_organizing_objects_where_needed)


   def _load_from_xml(self):
      root = self.root
      assert root is not None
      self.count = rqet.find_tag_int(root, 'Count')
      assert self.count > 0, 'empty grid connection set'
      self.cell_index_pairs_null_value = rqet.find_nested_tags_int(root, ['CellIndexPairs', 'NullValue'])
      self.face_index_pairs_null_value = rqet.find_nested_tags_int(root, ['LocalFacePerCellIndexPairs', 'NullValue'])
      # postpone loading of hdf5 array data till on-demand load (cell, grid & face index pairs)
      interp_root = rqet.find_tag(root, 'ConnectionInterpretations')
      if interp_root is None: return
      # load ordered feature list
      self.feature_list = []  # ordered list of (content_type, uuid, title) for faults
      for child in interp_root:
         if rqet.stripped_of_prefix(child.tag) != 'FeatureInterpretation': continue
         feature_type = rqet.content_type(rqet.find_tag_text(child, 'ContentType'))
         feature_uuid = bu.uuid_from_string(rqet.find_tag_text(child, 'UUID'))
         feature_title = rqet.find_tag_text(child, 'Title')
         # for now, only accept faults
         assert feature_type in ['obj_FaultInterpretation', 'obj_HorizonInterpretation']
         self.feature_list.append((feature_type, feature_uuid, feature_title))
         log.debug('connection set references fault interpretation: ' + feature_title)
      log.debug('number of faults referred to in connection set: ' + str(len(self.feature_list)))
      assert len(self.feature_list) > 0, 'list of fault interpretation references is empty for connection set'
      # leave feature indices till on-demand load
      self.grid_list = []
      for child in root:
         if rqet.stripped_of_prefix(child.tag) != 'Grid': continue
         grid_type = rqet.content_type(rqet.find_tag_text(child, 'ContentType'))
         grid_uuid = bu.uuid_from_string(rqet.find_tag_text(child, 'UUID'))
         assert grid_type == 'obj_IjkGridRepresentation', 'only IJK grids supported for grid connection sets'
         grid = self.model.grid(uuid = grid_uuid, find_properties = False)  # centralised list of grid objects for shared use
         self.grid_list.append(grid)
      if len(self.grid_list) == 0:  # this code only needed to handle defective datasets generated by earlier versions!
         log.warning('no grid nodes found in xml for connection set')
         grid = self.model.grid(find_properties = False)  # should find only or ROOT grid
         assert grid is not None, 'No ROOT grid found in model'
         self.grid_list = [grid]


   def set_pairs_from_kelp(self, kelp_0, kelp_1, feature_name, create_organizing_objects_where_needed, axis = 'K'):
      """Sets cell_index_pairs and face_index_pairs based on j and i face kelp strands, using simple no throw pairing."""

      # note: this method has been reworked to allow 'kelp' to be 'horizontal' strands when working in cross section

      if axis == 'K':
         kelp_k = None
         kelp_j = kelp_0
         kelp_i = kelp_1
      elif axis == 'J':
         kelp_k = kelp_0
         kelp_j = None
         kelp_i = kelp_1
      else:
         assert axis == 'I'
         kelp_k = kelp_0
         kelp_j = kelp_1
         kelp_i = None

      if feature_name is None: feature_name = 'feature from kelp lists'
      if len(self.grid_list) > 1:
         log.warning('setting grid connection set pairs from kelp for first grid in list only')
      grid = self.grid_list[0]
      if grid.nk > 1 and kelp_k is not None and len(kelp_k) > 0:
         if axis == 'J': k_layer = np.zeros((grid.nk - 1, grid.ni), dtype = bool)
         else: k_layer = np.zeros((grid.nk - 1, grid.nj), dtype = bool)
         kelp_a = np.array(kelp_k, dtype = int).T
         k_layer[kelp_a[0], kelp_a[1]] = True
         k_faces = np.zeros((grid.nk - 1, grid.nj, grid.ni), dtype = bool)
         if axis == 'J': k_faces[:] = k_layer.reshape((grid.nk - 1, 1, grid.ni))
         else: k_faces[:] = k_layer.reshape((grid.nk - 1, grid.nj, 1))
      else:
         k_faces = None
      if grid.nj > 1 and kelp_j is not None and len(kelp_j) > 0:
         if axis == 'K': j_layer = np.zeros((grid.nj - 1, grid.ni), dtype = bool)
         else: j_layer = np.zeros((grid.nk, grid.nj - 1), dtype = bool)
         kelp_a = np.array(kelp_j, dtype = int).T
         j_layer[kelp_a[0], kelp_a[1]] = True
         j_faces = np.zeros((grid.nk, grid.nj - 1, grid.ni), dtype = bool)
         if axis == 'K': j_faces[:] = j_layer.reshape((1, grid.nj - 1, grid.ni))
         else: j_faces[:] = j_layer.reshape((grid.nk, grid.nj - 1, 1))
      else:
         j_faces = None
      if grid.ni > 1 and kelp_i is not None and len(kelp_i) > 0:
         if axis == 'K': i_layer = np.zeros((grid.nj, grid.ni - 1), dtype = bool)
         else: i_layer = np.zeros((grid.nk, grid.ni - 1), dtype = bool)
         kelp_a = np.array(kelp_i, dtype = int).T
         i_layer[kelp_a[0], kelp_a[1]] = True
         i_faces = np.zeros((grid.nk, grid.nj, grid.ni - 1), dtype = bool)
         if axis == 'K': i_faces[:] = i_layer.reshape((1, grid.nj, grid.ni - 1))
         else: i_faces[:] = i_layer.reshape((grid.nk, 1, grid.ni - 1))
      else:
         i_faces = None
      self.set_pairs_from_face_masks(k_faces, j_faces, i_faces, feature_name, create_organizing_objects_where_needed)


   def set_pairs_from_face_masks(self, k_faces, j_faces, i_faces, feature_name, create_organizing_objects_where_needed,
                                 feature_type = 'fault'):   # other feature_type values: 'horizon', 'geobody boundary'
      """Sets cell_index_pairs and face_index_pairs based on triple face masks, using simple no throw pairing."""

      assert feature_type in ['fault', 'horizon', 'geobody boundary']
      if feature_name is None: feature_name = 'feature from face masks'     # not sure this default is wise
      if len(self.grid_list) > 1:
         log.warning('setting grid connection set pairs from face masks for first grid in list only')
      grid = self.grid_list[0]
      if feature_type == 'fault':
         feature_flavour = 'TectonicBoundaryFeature'
         interpretation_flavour = 'FaultInterpretation'
      else:
         feature_flavour = 'GeneticBoundaryFeature'
         if feature_type == 'horizon': interpretation_flavour = 'HorizonInterpretation'
         else: interpretation_flavour = 'GeobodyBoundaryInterpretation'
         # kind differentiates between horizon and geobody boundary
      fi_parts_list = self.model.parts_list_of_type(interpretation_flavour)
      if fi_parts_list is None or len(fi_parts_list) == 0:
         log.warning('no interpretation parts found in model for ' + feature_type)
      fi_uuid = None
      for fi_part in fi_parts_list:
         if self.model.title_for_part(fi_part).split()[0].lower() == feature_name.lower():
            fi_uuid = self.model.uuid_for_part(fi_part)
            break
      if fi_uuid is None:
         if create_organizing_objects_where_needed:
            tbf_parts_list = self.model.parts_list_of_type(feature_flavour)
            tbf = None
            for tbf_part in tbf_parts_list:
               if feature_name.lower() == self.model.title_for_part(tbf_part).split()[0].lower():
                  tbf_root = self.model.root_for_part(tbf_part)
                  if feature_type == 'fault':
                     tbf = rqo.TectonicBoundaryFeature(self.model, root_node = tbf_root)
                  else:
                     tbf = rqo.GeneticBoundaryFeature(self.model, kind = feature_type, root_node = tbf_root)
                  break
            if tbf is None:
               if feature_type == 'fault':
                  tbf = rqo.TectonicBoundaryFeature(self.model, kind = 'fault', feature_name = feature_name)
               else:
                  tbf = rqo.GeneticBoundaryFeature(self.model, kind = feature_type, feature_name = feature_name)
               tbf_root = tbf.create_xml()
            if feature_type == 'fault':
               fi = rqo.FaultInterpretation(self.model, tectonic_boundary_feature = tbf,
                                            is_normal = True)  # todo: set is_normal based on fault geometry in grid?
            elif feature_type == 'horizon':
               fi = rqo.HorizonInterpretation(self.model, genetic_boundary_feature = tbf)
               # todo: support boundary relation list and sequence stratigraphy surface
            else: # geobody boundary
               fi = rqo.GeobodyBoundaryInterpretation(self.model, genetic_boundary_feature = tbf)
            fi_root = fi.create_xml(tbf_root)
            fi_uuid = rqet.uuid_for_part_root(fi_root)
         else:
            log.error('no interpretation found for feature: ' + feature_name)
            return
      self.feature_list = [('obj_FaultInterpretation', fi_uuid, str(feature_name))]
      cell_pair_list = []
      face_pair_list = []
      nj_ni = grid.nj * grid.ni
      if k_faces is not None:
         for cell_kji0 in np.stack(np.where(k_faces)).T:
            cell = grid.natural_cell_index(cell_kji0)
            cell_pair_list.append((cell, cell + nj_ni))
            face_pair_list.append((self.face_index_map[0, 1], self.face_index_map[0, 0]))
      if j_faces is not None:
         for cell_kji0 in np.stack(np.where(j_faces)).T:
            cell = grid.natural_cell_index(cell_kji0)
            cell_pair_list.append((cell, cell + grid.ni))
            face_pair_list.append((self.face_index_map[1, 1], self.face_index_map[1, 0]))
      if i_faces is not None:
         for cell_kji0 in np.stack(np.where(i_faces)).T:
            cell = grid.natural_cell_index(cell_kji0)
            cell_pair_list.append((cell, cell + 1))
            face_pair_list.append((self.face_index_map[2, 1], self.face_index_map[2, 0]))
      self.cell_index_pairs = np.array(cell_pair_list, dtype = int)
      self.face_index_pairs = np.array(face_pair_list, dtype = int)
      self.count = len(self.cell_index_pairs)
      self.feature_indices = np.zeros(self.count, dtype = int)
      assert len(self.face_index_pairs) == self.count


   def set_pairs_from_faces_df(self, faces, create_organizing_objects_where_needed):
      """Sets cell_index_pairs and face_index_pairs based on pandas dataframe, using simple no throw pairing."""

      if len(self.grid_list) > 1:
         log.warning('setting grid connection set pairs from dataframe for first grid in list only')
      grid = self.grid_list[0]
      standardize_face_indicator_in_faces_df(faces)
      zero_base_cell_indices_in_faces_df(faces)
      faces = remove_external_faces_from_faces_df(faces, self.grid_list[0].extent_kji)
      self.feature_list = []
      cell_pair_list = []
      face_pair_list = []
      fi_list = []
      feature_index = 0
      name_list = faces['name'].unique()
      fi_parts_list = self.model.parts_list_of_type('FaultInterpretation')
      if fi_parts_list is None or len(fi_parts_list) == 0:
         log.warning('no fault interpretation parts found in model')
      fi_dict = {}  # maps fault name to interpretation uuid
      for fi_part in fi_parts_list:
         fi_dict[self.model.title_for_part(fi_part).split()[0].lower()] = self.model.uuid_for_part(fi_part)
      if create_organizing_objects_where_needed:
         tbf_parts_list = self.model.parts_list_of_type('TectonicBoundaryFeature')
      for name in name_list:
         # fetch uuid for fault interpretation object
         if name.lower() in fi_dict:
            fi_uuid = fi_dict[name.lower()]
         elif create_organizing_objects_where_needed:
            tbf = None
            for tbf_part in tbf_parts_list:
               if name.lower() == self.model.title_for_part(tbf_part).split()[0].lower():
                  tbf_root = self.model.root_for_part(tbf_part)
                  tbf = rqo.TectonicBoundaryFeature(self.model, root_node = tbf_root)
                  break
            if tbf is None:
               tbf = rqo.TectonicBoundaryFeature(self.model, feature_name = name)
               tbf_root = tbf.create_xml()
            fi = rqo.FaultInterpretation(self.model, tectonic_boundary_feature = tbf,
               is_normal = True)  # todo: set is_normal based on fault geometry in grid?
            fi_root = fi.create_xml(tbf_root)
            fi_uuid = rqet.uuid_for_part_root(fi_root)
            fi_dict[name.lower()] = fi_uuid
         else:
            log.error('no interpretation found for fault: ' + name)
            continue
         self.feature_list.append(('obj_FaultInterpretation', fi_uuid, str(name)))
         feature_faces = faces[faces['name'] == name]
         for i in range(len(feature_faces)):
            entry = feature_faces.iloc[i]
            f = entry['face']
            axis = 'KJI'.index(f[0])
            fp = '-+'.index(f[1])
            for k0 in range(entry['k1'], entry['k2'] + 1):
               for j0 in range(entry['j1'], entry['j2'] + 1):
                  for i0 in range(entry['i1'], entry['i2'] + 1):
                     neighbour = np.array([k0, j0, i0], dtype = int)
                     if fp: neighbour[axis] += 1
                     else: neighbour[axis] -= 1
                     fi_list.append(feature_index)
                     cell_pair_list.append((grid.natural_cell_index((k0, j0, i0)), grid.natural_cell_index(neighbour)))
                     face_pair_list.append((self.face_index_map[axis, fp], self.face_index_map[axis, 1 - fp]))
         feature_index += 1
      self.feature_indices = np.array(fi_list, dtype = int)
      self.cell_index_pairs = np.array(cell_pair_list, dtype = int)
      self.face_index_pairs = np.array(face_pair_list, dtype = int)
      self.count = len(self.cell_index_pairs)
      assert len(self.face_index_pairs) == self.count


   def append(self, other):
      """Adds the features in other grid connection set to this one."""

      # todo: check that grid is common to both connection sets
      assert len(self.grid_list) == 1 and len(other.grid_list) == 1, 'attempt to merge multi-grid connection sets'
      assert bu.matching_uuids(self.grid_list[0].uuid, other.grid_list[0].uuid),  \
            'attempt to merge grid connection sets from different grids'
      if self.count is None or self.count == 0:
         self.feature_list = other.feature_list.copy()
         self.count = other.count
         self.feature_indices = other.feature_indices.copy()
         self.cell_index_pairs = other.cell_index_pairs.copy()
         self.face_index_pairs = other.face_index_pairs.copy()
      else:
         feature_offset = len(self.feature_list)
         self.feature_list += other.feature_list
         combined_feature_indices = np.concatenate((self.feature_indices, other.feature_indices))
         combined_feature_indices[self.count:] += feature_offset
         combined_cell_index_pairs = np.concatenate((self.cell_index_pairs, other.cell_index_pairs))
         combined_face_index_pairs = np.concatenate((self.face_index_pairs, other.face_index_pairs))
         self.count += other.count
         self.cell_index_pairs = combined_cell_index_pairs
         self.face_index_pairs = combined_face_index_pairs
         self.feature_indices = combined_feature_indices
      self.uuid = bu.new_uuid()


   def single_feature(self, feature_index = None, feature_name = None):
      """Returns a new GridConnectionSet object containing the single feature copied from this set.

      note:
         the single feature connection set is established in memory but this method does not write
         to the hdf5 nor create xml or add as a part to the model

      :meta common:
      """

      assert feature_index is not None or feature_name is not None

      if feature_index is None: feature_index, _ = self.feature_index_and_uuid_for_fault_name(feature_name)
      if feature_index is None: return None
      if feature_index < 0 or feature_index >= len(self.feature_list): return None
      singleton = GridConnectionSet(self.model, grid = self.grid_list[0])
      singleton.cell_index_pairs, singleton.face_index_pairs =  \
         self.raw_list_of_cell_face_pairs_for_feature_index(feature_index)
      singleton.count = singleton.cell_index_pairs.shape[0]
      singleton.feature_indices = np.zeros((singleton.count,), dtype = int)
      singleton.feature_list = [self.feature_list[feature_index]]
      return singleton


   def filtered_by_layer_range(self, min_k0 = None, max_k0 = None):
      """Returns a new GridConnectionSet, being a copy with cell faces whittled down to a layer range.

      note:
         cells in layer max_k0 are included in the filtered set (not pythonesque)
      """

      self.cache_arrays()
      assert len(self.grid_list) == 1, 'attempt to filter multi-grid connection set by layer range'
      grid = self.grid_list[0]
      if min_k0 <= 0: min_k0 = None
      if max_k0 >= grid.extent_kji[0] - 1: max_k0 = None
      if min_k0 is None and max_k0 is None:
         dupe = GridConnectionSet(self.model, grid = grid)
         dupe.append(self)
         return dupe
      mask = np.zeros(grid.extent_kji, dtype = bool)
      if min_k0 is not None and max_k0 is not None:
         mask[min_k0:max_k0 + 1, :, :] = True
      elif min_k0 is not None:
         mask[min_k0:, :, :] = True
      else:
         mask[:max_k0 + 1, :, :] = True
      return self.filtered_by_cell_mask(mask)


   def filtered_by_cell_mask(self, mask, both_cells_required = True):
      """Returns a new GridConnectionSet, being a copy with cell faces whittled down by a boolean mask array."""

      assert len(self.grid_list) == 1, 'attempt to filter multi-grid connection set by cell mask'
      grid = self.grid_list[0]
      flat_extent = grid.cell_count()
      flat_mask = mask.reshape((flat_extent,))
      where_0 = flat_mask[self.cell_index_pairs[:, 0]]
      where_1 = flat_mask[self.cell_index_pairs[:, 1]]
      if both_cells_required:
         where_both = np.logical_and(where_0, where_1)
      else:
         where_both = np.logical_or(where_0, where_1)
      indices = np.where(where_both)[0]   # indices into primary axis of original arrays
      if len(indices) == 0:
         log.warning('no connections have passed filtering')
         return None
      masked_gcs = GridConnectionSet(self.model, grid = grid)
      masked_gcs.count = len(indices)
      masked_gcs.cell_index_pairs = self.cell_index_pairs[indices, :]
      masked_gcs.face_index_pairs = self.face_index_pairs[indices, :]
      masked_gcs.feature_indices = self.feature_indices[indices]
      masked_gcs.feature_list = self.feature_list.copy()
      masked_gcs.clean_feature_list()
      return masked_gcs


   def clean_feature_list(self):
      """Removes any features that have no associated connections."""

      keep_list = np.unique(self.feature_indices)
      if len(keep_list) == len(self.feature_list): return
      cleaned_list = []
      for i in range(len(keep_list)):
         assert i <= keep_list[i]
         if i != keep_list[i]:
            self.feature_indices[np.where(self.feature_indices == keep_list[i])[0]] = i
         cleaned_list.append(self.feature_list[keep_list[i]])
      self.feature_list = cleaned_list


   def cache_arrays(self):
      """Checks that the connection set array data is loaded and loads from hdf5 if not.

      :meta common:
      """

      if self.cell_index_pairs is None or self.face_index_pairs is None or (
         self.feature_list is not None and self.feature_indices is None):
         assert self.root is not None

      if self.cell_index_pairs is None:
         log.debug('caching cell index pairs from hdf5')
         cell_index_pairs_node = rqet.find_tag(self.root, 'CellIndexPairs')
         h5_key_pair = self.model.h5_uuid_and_path_for_node(cell_index_pairs_node, tag = 'Values')
         assert h5_key_pair is not None
         self.model.h5_array_element(h5_key_pair, cache_array = True, object = self,
                                     array_attribute = 'cell_index_pairs', dtype = 'int')

      if self.face_index_pairs is None:
         log.debug('caching face index pairs from hdf5')
         face_index_pairs_node = rqet.find_tag(self.root, 'LocalFacePerCellIndexPairs')
         h5_key_pair = self.model.h5_uuid_and_path_for_node(face_index_pairs_node, tag = 'Values')
         assert h5_key_pair is not None
         self.model.h5_array_element(h5_key_pair, cache_array = True, object = self,
                                     array_attribute = 'face_index_pairs', dtype = 'int32')

      if len(self.grid_list) > 1 and self.grid_index_pairs is None:
         grid_index_node = rqet.find_tag(self.root, 'GridIndexPairs')
         assert grid_index_node is not None, 'grid index pair data missing in grid connection set'
         h5_key_pair = self.model.h5_uuid_and_path_for_node(grid_index_node, tag = 'Values')
         assert h5_key_pair is not None
         self.model.h5_array_element(h5_key_pair, cache_array = True, object = self,
                                     array_attribute = 'grid_index_pairs', dtype = 'int')

      if self.feature_list is None: return
      interp_root = rqet.find_tag(self.root, 'ConnectionInterpretations')

      if self.feature_indices is None:
         log.debug('caching feature indices')
         elements_node = rqet.find_nested_tags(interp_root, ['InterpretationIndices', 'Elements'])
#         elements_node = rqet.find_nested_tags(interp_root, ['FaultIndices', 'Elements'])
         h5_key_pair = self.model.h5_uuid_and_path_for_node(elements_node, tag = 'Values')
         assert h5_key_pair is not None
         self.model.h5_array_element(h5_key_pair, cache_array = True, object = self,
                                     array_attribute = 'feature_indices', dtype = 'uint32')
         assert self.feature_indices.shape == (self.count,)

         cl_node = rqet.find_nested_tags(interp_root, ['InterpretationIndices', 'CumulativeLength'])
#         cl_node = rqet.find_nested_tags(interp_root, ['FaultIndices', 'CumulativeLength'])
         h5_key_pair = self.model.h5_uuid_and_path_for_node(cl_node, tag = 'Values')
         assert h5_key_pair is not None
         self.model.h5_array_element(h5_key_pair, cache_array = True, object = self,
                                     array_attribute = 'fi_cl', dtype = 'uint32')
         assert self.fi_cl.shape == (self.count,), 'connection set face pair(s) not assigned to exactly one fault'  # rough check
#        delattr(self, 'fi_cl')  # assumed to be one-to-one mapping, so cumulative length is discarded


   def number_of_grids(self):
      """Returns the number of grids involved in the connection set.

      :meta common:
      """

      return len(self.grid_list)


   def grid_for_index(self, grid_index):
      """Returns the grid object for the given grid_index."""

      assert 0 <= grid_index < len(self.grid_list)
      return self.grid_list[grid_index]


   def number_of_features(self):
      """Returns the number of features (faults) in the connection set.

      :meta common:
      """

      if self.feature_list is None: return 0
      return len(self.feature_list)


   def list_of_feature_names(self, strip = True):
      """Returns a list of the feature (fault) names for which the connection set has cell face data.

      :meta common:
      """

      if self.feature_list is None: return None
      name_list = []
      for (_, _, title) in self.feature_list:
         if strip:
            name_list.append(title.split()[0])
         else:
            name_list.append(title)
      return name_list


   def list_of_fault_names(self, strip = True):
      """Returns a list of the fault names for which the connection set has cell face data."""

      return self.list_of_feature_names(strip = strip)


   def feature_index_and_uuid_for_fault_name(self, fault_name):
      """Returns the index into the feature (fault) list for the named fault.

      arguments:
         fault_name (string): the name of the fault of interest

      returns:
         (index, uuid) where index is an integer which can be used to index the feature_list and is compatible
            with the elements of feature_indices (connection interpretations elements in xml);
            uuid is the uuid.UUID of the feature interpretation for the fault;
            (None, None) is returned if the named fault is not found in the feature list
      """

      if self.feature_list is None: return (None, None)
      # the protocol adopted here is that the citation title must start with the fault name
      index = 0
      for (_, uuid, title) in self.feature_list:
         if title.startswith(fault_name): return (index, uuid)
         index += 1
      return (None, None)


   def feature_name_for_feature_index(self, feature_index, strip = True):
      """Returns the fault name corresponding to the given index into the feature (fault) list.

      arguments:
         feature_index (non-negative integer): the index into the ordered feature list (fault interpretation list)
         strip (boolean, default True): if True, the citation title for the feature interpretation is split
            and the first word is returned (stripped of leading and trailing whitespace); if False, the
            citation title is returned unaltered

      returns:
         string being the citation title of the indexed fault interpretation part, optionally stripped
         down to the first word; by convention this is the name of the fault
      """

      if self.feature_list is None: return None
      if strip:
         return self.feature_list[feature_index][2].split()[0]
      return self.feature_list[feature_index][2]


   def fault_name_for_feature_index(self, feature_index, strip = True):
      """Returns the fault name corresponding to the given index into the feature (fault) list."""

      return self.feature_name_for_feature_index(feature_index = feature_index, strip = strip)


   def feature_index_for_cell_face(self, cell_kji0, axis, p01):
      """Returns the index into the feature (fault) list for the given face of the given cell, or None.

      note:
         where the cell face appears in more than one feature, the result will arbitrarily be the first
         occurrence in the cell_index_pair ordering of the grid connection set
      """

      self.cache_arrays()
      if self.feature_indices is None: return None
      cell = self.grid_list[0].natural_cell_index(cell_kji0)
      face = self.face_index_map[axis, p01]
      cell_matches = np.stack(np.where(self.cell_index_pairs == cell)).T
      for match in cell_matches[:]:
         if self.face_index_pairs[match[0], match[1]] == face: return self.feature_indices[match[0]]
      return None


   def raw_list_of_cell_face_pairs_for_feature_index(self, feature_index):
      """Returns list of cell face pairs contributing to feature (fault) with given index, in raw form.

      arguments:
         feature_index (non-negative integer): the index into the ordered feature list (fault interpretation list)

      returns:
         (cell_index_pairs, face_index_pairs) or (cell_index_pairs, face_index_pairs, grid_index_pairs)
         where each is an integer numpy array of shape (N, 2);
         if the connection set is for a single grid, the returned value is a pair, otherwise a triplet;
         the returned data is in raw form: normalized cell indices (for flattened array) and face indices
         in range 0..5; grid indices can be used to index the grid_list attribute for relevant Grid object
      """

      self.cache_arrays()
      if self.feature_indices is None: return None
      matches = np.where(self.feature_indices == feature_index)
      if len(self.grid_list) == 1:
         return self.cell_index_pairs[matches], self.face_index_pairs[matches]
      assert self.grid_index_pairs is not None
      return self.cell_index_pairs[matches], self.face_index_pairs[matches], self.grid_index_pairs[matches]


   def list_of_cell_face_pairs_for_feature_index(self, feature_index):
      """Returns list of cell face pairs contributing to feature (fault) with given index.

      arguments:
         feature_index (non-negative integer): the index into the ordered feature list (fault interpretation list)

      returns:
         (cell_index_pairs, face_index_pairs) or (cell_index_pairs, face_index_pairs, grid_index_pairs)
         where cell_index_pairs is a numpy int array of shape (N, 2, 3) being the paired kji0 cell indices;
         and face_index_pairs is a numpy int array of shape (N, 2, 2) being the paired face indices with the
         final axis of extent 2 indexed by 0 to give the facial axis (0 = K, 1 = J, 2 = I), and indexed by 1
         to give the facial polarity (0 = minus face, 1 = plus face);
         and grid_index_pairs is a numpy int array of shape (N, 2) the values of which can be used to index
         the grid_list attribute to access relevant Grid objects;
         if the connection set is for a single grid, the return value is a pair; otherwise a triplet
      """

      self.cache_arrays()
      if self.feature_indices is None: return None
      pairs_tuple = self.raw_list_of_cell_face_pairs_for_feature_index(feature_index)
      if len(self.grid_list) == 1:
         raw_cell_pairs, raw_face_pairs = pairs_tuple
         grid_pairs = None
      else:
         raw_cell_pairs, raw_face_pairs, grid_pairs = pairs_tuple

      cell_pairs = self.grid_list[0].denaturalized_cell_indices(raw_cell_pairs)
      face_pairs = self.face_index_inverse_map[raw_face_pairs]

      if grid_pairs is None: return cell_pairs, face_pairs
      return cell_pairs, face_pairs, grid_pairs


   def simplified_sets_of_kelp_for_feature_index(self, feature_index):
      """Returns a pair of sets of column indices, one for J+ faces, one for I+ faces, contributing to feature.

      arguments:
         feature_index (non-negative integer): the index into the ordered feature list (fault interpretation list)

      returns:
         (set of numpy pair of integers, set of numpy pair of integers) the first set is those columns where
         the J+ faces are contributing to the connection (or the J- faces of the neighbouring column); the second
         set is where the the I+ faces are contributing to the connection (or the I- faces of the neighbouring column)

      notes:
         K faces are ignored;
         this is compatible with the resqml baffle functionality elsewhere
      """

      cell_pairs, face_pairs = self.list_of_cell_face_pairs_for_feature_index(feature_index)
      simple_j_set = set()   # set of (j0, i0) pairs of column indices where J+ faces contribute, as 2 element numpy arrays
      simple_i_set = set()   # set of (j0, i0) pairs of column indices where I+ faces contribute, as 2 element numpy arrays
      for i in range(cell_pairs.shape[0]):
         for ip in range(2):
            cell_kji0 = cell_pairs[i, ip].copy()
            axis = face_pairs[i, ip, 0]
            if axis == 0: continue            # skip k faces
            polarity = face_pairs[i, ip, 1]
            if polarity == 0:
               cell_kji0[axis] -= 1
               if cell_kji0[axis] < 0: continue
            if axis == 1:  # J axis
               simple_j_set.add(tuple(cell_kji0[1:]))
            else:          # axis == 2; ie. I axis
               simple_i_set.add(tuple(cell_kji0[1:]))
      return (simple_j_set, simple_i_set)


   def rework_face_pairs(self):
      """Overwrites the in-memory array of face pairs to reflect neighbouring I or J columns.

      note:
         the indexing of faces within a cell is not clearly documented in the RESQML guides;
         the constant self.face_index_map and its inverse is the best guess at what this mapping is;
         this function overwrites the faces within cell data where a connected pair are in neighbouring
         I or J columns, using the face within cell values that fit with the neighbourly relationship;
         for some implausibly extreme gridding, this might not give the correct result
      """

      self.cache_arrays()
      if self.feature_indices is None: return None

      assert len(self.grid_list) == 1

      cell_pairs = self.grid_list[0].denaturalized_cell_indices(self.cell_index_pairs)

      modifications = 0
      for i in range(self.count):
         kji0_pair = cell_pairs[i]
         new_faces = None
         if kji0_pair[0][1] == kji0_pair[1][1]:   # same J index
            if kji0_pair[0][2] == kji0_pair[1][2] + 1:
               new_faces = (self.face_index_map[2, 0], self.face_index_map[2, 1])
            elif kji0_pair[0][2] == kji0_pair[1][2] - 1:
               new_faces = (self.face_index_map[2, 1], self.face_index_map[2, 0])
         elif kji0_pair[0][2] == kji0_pair[1][2]: # same I index
            if kji0_pair[0][1] == kji0_pair[1][1] + 1:
               new_faces = (self.face_index_map[1, 0], self.face_index_map[1, 1])
            elif kji0_pair[0][1] == kji0_pair[1][1] - 1:
               new_faces = (self.face_index_map[1, 1], self.face_index_map[1, 0])
         if new_faces is not None and np.any(self.face_index_pairs[i] != new_faces):
             self.face_index_pairs[i] = new_faces
             modifications += 1
      log.debug('number of face pairs modified during neighbour based reworking: ' + str(modifications))


   def write_hdf5(self, file_name = None, mode = 'a'):
      """Create or append to an hdf5 file, writing datasets for the connection set after caching arrays.

      :meta common:
      """

      # NB: cell pairs, face pairs, and feature indices arrays must be ready, in memory
      # note: grid pairs not yet supported
      # prepare one-to-one mapping of cell face pairs with feaature indices
      if not file_name: file_name = self.model.h5_file_name()
      log.debug('gcs write to hdf5 file: ' + file_name + ' mode: ' + mode)

      one_to_one = np.empty(self.feature_indices.shape, dtype = int)
      for i in range(one_to_one.size): one_to_one[i] = i + 1

      h5_reg = rwh5.H5Register(self.model)

      # register arrays
      # note: this implementation requires each cell face pair to be assigned to exactly one interpretation
      # uuid/CellIndexPairs  (N, 2)  int64
      h5_reg.register_dataset(self.uuid, 'CellIndexPairs', self.cell_index_pairs)
      if len(self.grid_list) > 1 and self.grid_index_pairs is not None:
         # uuid/GridIndexPairs  (N, 2)  int64
         h5_reg.register_dataset(self.uuid, 'GridIndexPairs', self.grid_index_pairs)
      # uuid/LocalFacePerCellIndexPairs  (N, 2)  int32
      h5_reg.register_dataset(self.uuid, 'LocalFacePerCellIndexPairs', self.face_index_pairs)
      # uuid/FaultIndices/elements  (N,)  uint32
      h5_reg.register_dataset(self.uuid, 'InterpretationIndices/elements', self.feature_indices)
#      h5_reg.register_dataset(self.uuid, 'FaultIndices/elements', self.feature_indices)
      # uuid/FaultIndices/cumulativeLength  (N,)  uint32
      h5_reg.register_dataset(self.uuid, 'InterpretationIndices/cumulativeLength', one_to_one)
#      h5_reg.register_dataset(self.uuid, 'FaultIndices/cumulativeLength', one_to_one)

      h5_reg.write(file_name, mode = mode)


   def create_xml(self, ext_uuid = None, add_as_part = True, add_relationships = True, title = None, originator = None):
      """Creates a Grid Connection Set (fault faces) xml node and optionally adds as child of root and/or to parts forest.

      :meta common:
      """

      # NB: only one grid handled for now
      # xml for grid(s) must be created before calling this method

      if ext_uuid is None: ext_uuid = self.model.h5_uuid()
      if not self.title and not title: title = 'ROOT'

      gcs = super().create_xml(add_as_part = False, title = title, originator = originator)

      c_node = rqet.SubElement(gcs, ns['resqml2'] + 'Count')
      c_node.set(ns['xsi'] + 'type', ns['xsd'] + 'positiveInteger')
      c_node.text = str(self.count)

      cip_node = rqet.SubElement(gcs, ns['resqml2'] + 'CellIndexPairs')
      cip_node.set(ns['xsi'] + 'type', ns['resqml2'] + 'IntegerHdf5Array')
      cip_node.text = '\n'

      cip_null = rqet.SubElement(cip_node, ns['resqml2'] + 'NullValue')
      cip_null.set(ns['xsi'] + 'type', ns['xsd'] + 'integer')
      cip_null.text = str(self.cell_index_pairs_null_value)

      cip_values = rqet.SubElement(cip_node, ns['resqml2'] + 'Values')
      cip_values.set(ns['xsi'] + 'type', ns['eml'] + 'Hdf5Dataset')
      cip_values.text = '\n'

      self.model.create_hdf5_dataset_ref(ext_uuid, self.uuid, 'CellIndexPairs', root = cip_values)

      if len(self.grid_list) > 1 and self.grid_index_pairs is not None:

         gip_node = rqet.SubElement(gcs, ns['resqml2'] + 'GridIndexPairs')
         gip_node.set(ns['xsi'] + 'type', ns['resqml2'] + 'IntegerHdf5Array')
         gip_node.text = '\n'

         gip_null = rqet.SubElement(gip_node, ns['resqml2'] + 'NullValue')
         gip_null.set(ns['xsi'] + 'type', ns['xsd'] + 'integer')
         gip_null.text = str(self.grid_index_pairs_null_value)

         gip_values = rqet.SubElement(gip_node, ns['resqml2'] + 'Values')
         gip_values.set(ns['xsi'] + 'type', ns['eml'] + 'Hdf5Dataset')
         gip_values.text = '\n'

         self.model.create_hdf5_dataset_ref(ext_uuid, self.uuid, 'GridIndexPairs', root = gip_values)

      fip_node = rqet.SubElement(gcs, ns['resqml2'] + 'LocalFacePerCellIndexPairs')
      fip_node.set(ns['xsi'] + 'type', ns['resqml2'] + 'IntegerHdf5Array')
      fip_node.text = '\n'

      fip_null = rqet.SubElement(fip_node, ns['resqml2'] + 'NullValue')
      fip_null.set(ns['xsi'] + 'type', ns['xsd'] + 'integer')
      fip_null.text = str(self.face_index_pairs_null_value)

      fip_values = rqet.SubElement(fip_node, ns['resqml2'] + 'Values')
      fip_values.set(ns['xsi'] + 'type', ns['eml'] + 'Hdf5Dataset')
      fip_values.text = '\n'

      self.model.create_hdf5_dataset_ref(ext_uuid, self.uuid, 'LocalFacePerCellIndexPairs', root = fip_values)

      if self.feature_indices is not None and self.feature_list is not None:

         ci_node = rqet.SubElement(gcs, ns['resqml2'] + 'ConnectionInterpretations')
         ci_node.set(ns['xsi'] + 'type', ns['resqml2'] + 'ConnectionInterpretations')
         ci_node.text = '\n'

         ii = rqet.SubElement(ci_node, ns['resqml2'] + 'InterpretationIndices')
#         ii = rqet.SubElement(ci_node, ns['resqml2'] + 'FaultIndices')
         ii.set(ns['xsi'] + 'type', ns['resqml2'] + 'ResqmlJaggedArray')
         ii.text = '\n'

         elements = rqet.SubElement(ii, ns['resqml2'] + 'Elements')
         elements.set(ns['xsi'] + 'type', ns['resqml2'] + 'IntegerHdf5Array')
         elements.text = '\n'

         el_null = rqet.SubElement(elements, ns['resqml2'] + 'NullValue')
         el_null.set(ns['xsi'] + 'type', ns['xsd'] + 'integer')
         el_null.text = '-1'

         el_values = rqet.SubElement(elements, ns['resqml2'] + 'Values')
         el_values.set(ns['xsi'] + 'type', ns['eml'] + 'Hdf5Dataset')
         el_values.text = '\n'

         self.model.create_hdf5_dataset_ref(ext_uuid, self.uuid, 'InterpretationIndices/elements', root = el_values)
#         self.model.create_hdf5_dataset_ref(ext_uuid, self.uuid, 'FaultIndices/elements', root = el_values)

         c_length = rqet.SubElement(ii, ns['resqml2'] + 'CumulativeLength')
         c_length.set(ns['xsi'] + 'type', ns['resqml2'] + 'IntegerHdf5Array')
         c_length.text = '\n'

         cl_null = rqet.SubElement(c_length, ns['resqml2'] + 'NullValue')
         cl_null.set(ns['xsi'] + 'type', ns['xsd'] + 'integer')
         cl_null.text = '-1'

         cl_values = rqet.SubElement(c_length, ns['resqml2'] + 'Values')
         cl_values.set(ns['xsi'] + 'type', ns['eml'] + 'Hdf5Dataset')
         cl_values.text = '\n'

         self.model.create_hdf5_dataset_ref(ext_uuid, self.uuid, 'InterpretationIndices/cumulativeLength', root = cl_values)
#         self.model.create_hdf5_dataset_ref(ext_uuid, self.uuid, 'FaultIndices/cumulativeLength', root = cl_values)

         # add feature interpretation reference node for each fault in list, NB: ordered list
         for (f_content_type, f_uuid, f_title) in self.feature_list:

            # for now, only support connection sets for faults
            if f_content_type == 'obj_FaultInterpretation':
               fi_part = rqet.part_name_for_object('obj_FaultInterpretation', f_uuid)
               fi_root = self.model.root_for_part(fi_part)
               self.model.create_ref_node('FeatureInterpretation', self.model.title_for_root(fi_root),
                                    f_uuid, content_type = 'obj_FaultInterpretation', root = ci_node)
            elif f_content_type == 'obj_HorizonInterpretation':
               fi_part = rqet.part_name_for_object('obj_HorizonInterpretation', f_uuid)
               fi_root = self.model.root_for_part(fi_part)
               self.model.create_ref_node('FeatureInterpretation', self.model.title_for_root(fi_root),
                                    f_uuid, content_type = 'obj_HorizonInterpretation', root = ci_node)
            else:
               raise Exception('unsupported content type in grid connection set')

      for grid in self.grid_list:
         self.model.create_ref_node('Grid', self.model.title_for_root(grid.root),
                                    grid.uuid, content_type = 'obj_IjkGridRepresentation', root = gcs)

      if add_as_part:
         self.model.add_part('obj_GridConnectionSetRepresentation', self.uuid, gcs)

         if add_relationships:
            for (obj_type, f_uuid, _) in self.feature_list:
               fi_part = rqet.part_name_for_object(obj_type, f_uuid)
               fi_root = self.model.root_for_part(fi_part)
               self.model.create_reciprocal_relationship(gcs, 'destinationObject', fi_root, 'sourceObject')
            ext_part = rqet.part_name_for_object('obj_EpcExternalPartReference', ext_uuid, prefixed = False)
            ext_node = self.model.root_for_part(ext_part)
            self.model.create_reciprocal_relationship(gcs, 'mlToExternalPartProxy', ext_node, 'externalPartProxyToMl')
            for grid in self.grid_list:
               self.model.create_reciprocal_relationship(gcs, 'destinationObject', grid.root, 'sourceObject')

      return gcs


   def write_simulator(self, filename, mode = 'w', simulator = 'nexus', include_both_sides = False, use_minus = False):
      """Creates a Nexus include file holding FAULTS (or MULT) keyword and data."""

      assert simulator == 'nexus'

      active_nexus_head = None
      row_count = 0

      def write_nexus_header_lines(fp, axis, polarity, fault_name, grid_name = 'ROOT'):
         nonlocal active_nexus_head
         if (axis, polarity, fault_name, grid_name) == active_nexus_head: return
         T = 'T' + 'ZYX'[axis]
         plus_minus = ['MINUS', 'PLUS'][polarity]
         fp.write('\nMULT\t' + T + '\tALL\t' + plus_minus + '\tMULT\n')
         fp.write('\tGRID\t' + grid_name + '\n')
         fp.write('\tFNAME\t' + fault_name + '\n')
         if len(fault_name) > 256:
            log.warning('exported fault name longer than Nexus limit of 256 characters: ' + fault_name)
            tm.log_nexus_tm('warning')
         active_nexus_head = (axis, polarity, fault_name, grid_name)

      def write_row(gcs, fp, name, i, j, k1, k2, axis, polarity):
         nonlocal row_count
         write_nexus_header_lines(fp, axis, polarity, name)
         fp.write('\t{0:1d}\t{1:1d}\t{2:1d}\t{3:1d}\t{4:1d}\t{5:1d}\t1.0\n'.format(
                  i + 1, i + 1, j + 1, j + 1, k1 + 1, k2 + 1))
         row_count += 1

      log.info('writing fault data in simulator format to file: ' + filename)

      if include_both_sides: sides = [0, 1]
      elif use_minus: sides = [1]
      else: sides = [0]

      with open(filename, mode, newline = '') as fp:
         for feature_index in range(len(self.feature_list)):
            feature_name = self.feature_list[feature_index][2].split()[0].upper()
            cell_index_pairs, face_index_pairs = self.list_of_cell_face_pairs_for_feature_index(feature_index)
            for side in sides:
               both = np.empty((cell_index_pairs.shape[0], 5), dtype = int)
               both[:, :2] = face_index_pairs[:, side, :]  # axis, polarity
               both[:, 2:] = cell_index_pairs[:, side, :]  # k, j, i
               df = pd.DataFrame(both, columns = ['axis', 'polarity', 'k', 'j', 'i'])
               df = df.sort_values(by = ['axis', 'polarity', 'j', 'i', 'k'])
               both_sorted = np.empty(both.shape, dtype = int)
               both_sorted[:] = df
               cell_indices = both_sorted[:, 2:]
               face_indices = np.empty((both_sorted.shape[0], 2), dtype = int)
               face_indices[:, :] = both_sorted[:, :2]
               del both_sorted
               del both
               del df
               k = None
               i = j = k2 = axis = polarity = None # only needed to placate flake8 which whinges incorrectly otherwise
               for row in range(cell_indices.shape[0]):
                  kp, jp, ip = cell_indices[row]
                  axis_p, polarity_p = face_indices[row]
                  if k is not None:
                     if axis_p != axis or polarity_p != polarity or ip != i or jp != j or kp != k2 + 1:
                        write_row(self, fp, feature_name, i, j, k, k2, axis, polarity)
                        k = None
                     else:
                        k2 = kp
                  if k is None:
                     i = ip
                     j = jp
                     k = k2 = kp
                     axis = axis_p
                     polarity = polarity_p
               if k is not None:
                  write_row(self, fp, feature_name, i, j, k, k2, axis, polarity)


   def get_column_edge_list_for_feature(self,feature,gridindex=0,min_k=0,max_k=0):
      """Extracts a list of cell faces for a given feature index, over a given range of layers in the grid

      Args:
         feature - feature index
         gridindex - index of grid to be used in grid connection set gridlist, default 0
         min_k - minimum k layer, default 0
         max_k - maximum k layer, default 0
      Returns:
         list of cell faces for the feature (j_col, i_col, face_axis, face_polarity)
      """
      subgcs = self.filtered_by_layer_range(min_k0=min_k, max_k0=max_k)

      cell_face_details = subgcs.list_of_cell_face_pairs_for_feature_index(feature)

      if len(cell_face_details) == 2:
         assert gridindex == 0, 'Only one grid present'
         cell_ind_pairs, face_ind_pairs = cell_face_details
         grid_ind_pairs = None
      else:
         cell_ind_pairs, face_ind_pairs, grid_ind_pairs = cell_face_details
      ij_faces = []
      for i in range(len(cell_ind_pairs)):
         for a_or_b in [0,1]:
            if grid_ind_pairs is None or grid_ind_pairs[i,a_or_b] == gridindex:
                if face_ind_pairs[i,a_or_b,0] != 0:
                   entry = ((int(cell_ind_pairs[i,a_or_b,1]),
                             int(cell_ind_pairs[i,a_or_b,2]),
                             int(face_ind_pairs[i,a_or_b,0])-1, # in the outputs j=0, i=1
                             int(face_ind_pairs[i,a_or_b,1])))  # in the outputs negativeface=0, positiveface=1
                   if entry in ij_faces: continue
                   ij_faces.append(entry)
      ij_faces_np = np.array((ij_faces))
      return ij_faces_np


   def get_column_edge_bool_array_for_feature(self,feature,gridindex=0,min_k=0,max_k=0):
      """Generate a boolean aray defining which column edges are present for a given feature and k-layer range

      Args:
         feature - feature index
         gridindex - index of grid to be used in grid connection set gridlist, default 0
         min_k - minimum k layer
         max_k - maximum k layer
      Returns:
         boolean fault_by_column_edge_mask array (shape nj,ni,2,2)

      Note: the indices for the final two axes define the edges:
         the first defines j or i (0 or 1)
         the second negative or positive face (0 or 1)

         so [[True,False],[False,True]] indicates the -j and +i edges of the column are present
      """
      cell_face_list = self.get_column_edge_list_for_feature(feature,gridindex,min_k,max_k)

      fault_by_column_edge_mask = np.zeros((self.grid_list[gridindex].nj,self.grid_list[gridindex].ni,2,2),dtype=bool)
      for i in cell_face_list:  fault_by_column_edge_mask[tuple(i)] = True

      return fault_by_column_edge_mask


   def get_property_by_feature_index_list(self, feature_index_list = None, property_name = 'Transmissibility multiplier'):
      """Returns a list of property values by feature.

      arguments:
         feature_index_list (list of int, optional): if present, the feature indices for which property values will be included
            in the resulting list; if None, values will be included for each feature in the feature_list for this connection set
         property_name (string, default 'Transmissibility multiplier'): the property name of interest, as used in the features'
            extra metadata as a key

      returns:
         list of float being the list of property values for the list of features, in corresponding order
      """

      if feature_index_list is None: feature_index_list = range(self.feature_list)
      value_list = []
      for feature_index in feature_index_list:
         _, feature_uuid, _ = self.feature_list[feature_index]
         feat = rqo.FaultInterpretation(parent_model = self.model,
                                        root_node = self.model.root(uuid = feature_uuid))
         if property_name not in feat.extra_metadata.keys(): log.info(f'Property name {property_name} not found in extra_metadata for {self.model.citation_title_for_part(self.model.part_for_uuid(feature_uuid))}')
         else: value_list.append(float(feat.extra_metadata[property_name]))
      return value_list


   def get_column_edge_float_array_for_feature(self, feature, fault_by_column_edge_mask, property_name='Transmissibility multiplier'):
      """Generate a float value aray defining the property valyes for different column edges present for a given feature and k-layer range

      Args:
         feature - feature index
         fault_by_column_edge_mask - fault_by_column_edge_mask with True on edges where feature is present
         property_name - name of property, should be present within the FaultInterpreation feature metadata
      Returns:
         float property_value_by_column_edge array (shape nj,ni,2,2)

      Note: the indices for the final two axes define the edges:
         the first defines j or i (0 or 1)
         the second negative or positive face (0 or 1)

         so [[1,np.nan],[np.nan,np.nan]] indicates the -j edge of the column are present with a value of 1
      """

      prop_values = self.get_property_by_feature_index_list(feature_index_list = [feature], property_name = property_name)
      if prop_values == []: return None
      property_value_by_column_edge = np.where(fault_by_column_edge_mask, prop_values[0], np.nan)
      return property_value_by_column_edge


   def get_combined_fault_mask_index_value_arrays(self,gridindex=0,min_k=0,max_k=0,
                                                  property_name='Transmissibility multiplier',feature_list=None):
      """Generate a combined mask, index and value arrays for all column edges across a k-layer range, for a defined feature_list

      Args:
         gridindex - index of grid to be used in grid connection set gridlist, default 0
         min_k - minimum k_layer
         max_k - maximum k_layer
         property_name - name of property, should be present within the FaultInterpreation feature metadata
         feature_list - list of feature index numbers to run for, default to all features
      Returns:
         bool array mask showing all column edges within features (shape nj,ni,2,2)
         int array showing the feature index for all column edges within features (shape nj,ni,2,2)
         float array showing the property value for all column edges within features (shape nj,ni,2,2)
      """
      self.cache_arrays()
#     if feature_list is None: feature_list = np.unique(self.feature_indices)
      if feature_list is None: feature_list = range(len(self.feature_list))
      sum_unmasked = None

      for i, feature in enumerate(feature_list):
         fault_by_column_edge_mask = self.get_column_edge_bool_array_for_feature(feature,gridindex,min_k=min_k,max_k=max_k)
         property_value_by_column_edge = self.get_column_edge_float_array_for_feature(feature, fault_by_column_edge_mask, property_name=property_name)
         if i == 0:
            combined_mask = fault_by_column_edge_mask.copy()
            if property_value_by_column_edge is not None:
               combined_values = property_value_by_column_edge.copy()
            else: combined_values = None
            combined_index = np.full((fault_by_column_edge_mask.shape),-1,dtype=int)
            combined_index = np.where(fault_by_column_edge_mask, feature, combined_index)
            sum_unmasked = np.sum(fault_by_column_edge_mask)
         else:
            combined_mask = np.logical_or(combined_mask,fault_by_column_edge_mask)
            if property_value_by_column_edge is not None:
               if combined_values is not None:
                  combined_values = np.where(fault_by_column_edge_mask, property_value_by_column_edge, combined_values)
               else: combined_values = property_value_by_column_edge.copy()
            combined_index = np.where(fault_by_column_edge_mask, feature, combined_index)
            sum_unmasked += np.sum(fault_by_column_edge_mask)

      if not np.sum(combined_mask) == sum_unmasked: log.warning("One or more features exist across the same column edge!")
      return combined_mask, combined_index, combined_values


   def tr_property_array(self, fa = None, tol_fa = 0.0001, tol_half_t = 1.0e-5, apply_multipliers = False):
      """Return a transmissibility array with one value per pair in the connection set.

      argument:
         fa (numpy float array of shape (count, 2), optional): if present, the fractional area for each pair connection,
            from perspective of each side of connection; if None, a fractional area of one is assumed
         tol_fa (float, default 0.0001): fractional area tolerance – if the fractional area on either side of a
            juxtaposition is less than this, then the corresponding transmissibility is set to zero
         tol_half_t (float, default 1.0e-5): if the half cell transmissibility either side of a juxtaposition is
            less than this, the corresponding transmissibility is set to zero; units are as for returned values (see
            notes)
         apply_multipliers (boolean, default False): if True, a transmissibility multiplier for each feature is
            extracted from the feature extra metadata and applied to the transmissibility calculation

      returns:
         numpy float array of shape (count,) being the absolute transmissibilities across the connected cell face pairs;
         see notes regarding units

      notes:
         implicit units of measure of returned values will be m3.cP/(kPa.d) if grids' crs length units are metres,
         bbl.cP/(psi.d) if length units are feet; the computation is compatible with the Nexus NEWTRAN formulation;
         multiple grids are assumed to be in the same units and z units must be the same as xy units
      """

      feature_mult_list = self.get_property_by_feature_index_list() if apply_multipliers else None

      count = self.count
      if fa is not None: assert fa.shape == (count, 2)
      f_tr = np.zeros(count)
      half_t_list = []
      for grid in self.grid_list:
         half_t_list.append(grid.half_cell_transmissibility())  # todo: pass realization argument
      single_grid = (self.number_of_grids() == 1)
      kji0 = self.grid_list[0].denaturalized_cell_indices(self.cell_index_pairs) if single_grid else None
      # kji0 shape (count, 2, 3); 2 being m,p; 3 being k,j,i; multi-grid cell indices denaturalised within for loop below
      fa_m = fa_p = 1.0
      gi_m = gi_p = 0  # used below in single grid case
      for e in range(count):
         axis_m, polarity_m = self.face_index_inverse_map[self.face_index_pairs[e, 0]]
         axis_p, polarity_p = self.face_index_inverse_map[self.face_index_pairs[e, 0]]
         if single_grid:
            kji0_m = kji0[e, 0]
            kji0_p = kji0[e, 1]
         else:
            gi_m = self.grid_index_pairs[e, 0]
            gi_p = self.grid_index_pairs[e, 1]
            kji0_m = self.grid_list[gi_m].denaturalized_cell_index(self.cell_index_pairs[e, 0])
            kji0_p = self.grid_list[gi_p].denaturalized_cell_index(self.cell_index_pairs[e, 1])
         half_t_m = half_t_list[gi_m][kji0_m[0], kji0_m[1], kji0_m[2], axis_m, polarity_m]
         half_t_p = half_t_list[gi_p][kji0_p[0], kji0_p[1], kji0_p[2], axis_p, polarity_p]
         if half_t_m < tol_half_t or half_t_p < tol_half_t: continue
         if fa is not None:
            fa_m, fa_p = fa[e]
            if fa_m < tol_fa or fa_p < tol_fa: continue
         mult_tr = feature_mult_list[self.feature_indices[e]] if apply_multipliers else 1.0
         f_tr[e] = mult_tr / (1.0 / (half_t_m * fa_m) +  1.0 / (half_t_p * fa_p))
      return f_tr


   def inherit_features(self, featured):
      """Inherit features from another connection set, reassigning feature indices for matching faces.

      arguments:
         featured (GridConnectionSet): the other connection set which holds features to be inherited

      notes:
         the list of cell face pairs in this set remains unchanged; the corresponding feature indices
         are updated based on individual cell faces that are found in the featured connection set;
         the feature list is extended with inherited features as required; this method is typically
         used to populate a fault connection set built from grid geometry with named fault features
         from a simplistic connection set, thus combining full geometric juxtaposition information
         with named features; currently restricted to single grid connection sets
      """

      def sorted_paired_cell_face_index_position(cell_face_index, a_or_b):
         # pair one side (a_or_b) of cell_face_index with its position, then sort
         count = len(cell_face_index)
         sp = np.empty((count, 2), dtype = int)
         sp[:, 0] = cell_face_index[:, a_or_b]
         sp[:, 1] = np.arange(count)
         t = [tuple(r) for r in sp]  # could use numpy fields based sort instead of tuple list?
         t.sort()
         return np.array(t)

      def find_in_sorted(paired, v):

         def fis(p, v, a, b):  # recursive binary search
            if a >= b: return None
            m = (a + b) // 2
            s = p[m, 0]
            if s == v: return p[m, 1]
            if s > v: return fis(p, v, a, m)
            return fis(p, v, m + 1, b)

         return fis(paired, v, 0, len(paired))

      assert len(self.grid_list) == 1 and len(featured.grid_list) == 1
      assert bu.matching_uuids(self.grid_list[0].uuid, featured.grid_list[0].uuid)

      # build combined feature list and mapping of feature indices
      featured.cache_arrays()
      original_feature_count = len(self.feature_list)
      featured_index_map = []  # maps from feature index in featured to extended feature list in this set
      feature_uuid_list = [bu.uuid_from_string(u) for _, u, _ in self.feature_list]  # bu call probably not needed
      for featured_triplet in featured.feature_list:
         featured_uuid = bu.uuid_from_string(featured_triplet[1])  # bu call probably not needed
         if featured_uuid in feature_uuid_list:
            featured_index_map.append(feature_uuid_list.index[featured_uuid])
         else:
            featured_index_map.append(len(self.feature_list))
            self.feature_list.append(featured_triplet)
            feature_uuid_list.append(featured_uuid)
            self.model.copy_part_from_other_model(featured, featured.model.part(uuid = featured_uuid))

      # convert cell index, face index to a single integer to facilitate sorting and comparison
      cell_face_index_self = self.compact_indices()
      cell_face_index_featured = featured.compact_indices()

      # sort all 4 (2 sides in each of 2 sets) cell_face index data, keeping track of positions in original lists
      cfp_a_self = sorted_paired_cell_face_index_position(cell_face_index_self, 0)
      cfp_b_self = sorted_paired_cell_face_index_position(cell_face_index_self, 1)
      cfp_a_featured = sorted_paired_cell_face_index_position(cell_face_index_featured, 0)
      cfp_b_featured = sorted_paired_cell_face_index_position(cell_face_index_featured, 1)

      # for each cell face in self, look for same in featured and inherit feature index if found
      previous_cell_face_index = previous_feature_index = None
      for (a_or_b, cfp_self) in [(0, cfp_a_self), (1, cfp_b_self)]:  # could risk being lazy and only using one side?
         for (cell_face_index, place) in cfp_self:
            if self.feature_indices[place] >= original_feature_count: continue  # already been updated
            if cell_face_index == previous_cell_face_index:
               if previous_feature_index is not None:
                  self.feature_indices[place] = previous_feature_index
            elif cell_face_index == self.face_index_pairs_null_value:
               continue
            else:
               featured_place = find_in_sorted(cfp_a_featured, cell_face_index)
               if featured_place is None:
                  featured_place = find_in_sorted(cfp_b_featured, cell_face_index)
               if featured_place is None:
                  previous_feature_index = None
               else:
                  featured_index = featured.feature_indices[featured_place]
                  self.feature_indices[place] = previous_feature_index = featured_index_map[featured_index]
               previous_cell_face_index = cell_face_index

      # clean up by removing any original features no longer in use
      removals = []
      for fi in range(original_feature_count):
         if fi not in self.feature_indices: removals.append(fi)
      reduction = len(removals)
      if reduction > 0:
         while len(removals) > 0:
            self.feature_list.pop(removals[-1])
            removals.pop()
         self.feature_indices[:] -= reduction


   def compact_indices(self):
      """Returns numpy int array of shape (count, 2) combining each cell index, face index into a single integer."""
      if self.cell_index_pairs is None or self.face_index_pairs is None: return None
      null_mask = np.logical_or(self.cell_index_pairs == self.cell_index_pairs_null_value,
                                self.face_index_pairs == self.face_index_pairs_null_value)
      combined = 6 * self.cell_index_pairs + self.face_index_pairs
      return np.where(null_mask, self.face_index_pairs_null_value, combined)


def pinchout_connection_set(grid, skip_inactive = True, feature_name = 'pinchout'):
   """Returns a new GridConnectionSet representing non-standard K face connections across pinchouts.

   arguments:
      grid (grid.Grid): the grid for which a pinchout connection set is required
      skip_inactive (boolean, default True): if True, connections are not included where there is an inactive cell
         above or below the pinchout; if False, such connections are included
      feature_name (string, default 'pinchout'): the name to use as citation title in the feature and interpretation

   notes:
      this function does not write to hdf5, nor create xml for the new grid connection set;
      however, it does create one feature and a corresponding interpretation and creates xml for those
   """

   assert grid is not None

   po = grid.pinched_out()
   dead = grid.extract_inactive_mask() if skip_inactive else None

   cip_list = []  # cell index pair list

   for j in range(grid.nj):
      for i in range(grid.ni):
         ka = 0
         while True:
            while ka < grid.nk - 1 and po[ka, j, i]: ka += 1
            while ka < grid.nk - 1 and not po[ka + 1, j, i]: ka += 1
            if ka >= grid.nk - 1: break
            # ka now in non-pinched out cell above pinchout
            if (skip_inactive and dead[ka, j, i]) or (grid.k_gaps and grid.k_gap_after[ka]):
               ka += 1
               continue
            kb = ka + 1
            while kb < grid.nk and po[kb, j, i]: kb += 1
            if kb >= grid.nk: break
            if skip_inactive and dead[kb, j, i]:
               ka = kb + 1
               continue
            # kb now beneath pinchout
            cip_list.append((grid.natural_cell_index((ka, j, i)), grid.natural_cell_index((kb, j, i))))
            ka = kb + 1

   log.debug(f'{len(cip_list)} pinchout connections found')

   pcs = _make_k_gcs_from_cip_list(grid, cip_list, feature_name)

   return pcs



def k_gap_connection_set(grid, skip_inactive = True, feature_name = 'k gap connection', tolerance = 0.001):
   """Returns a new GridConnectionSet representing K face connections where a K gap is zero thickness.

   arguments:
      grid (grid.Grid): the grid for which a K gap connection set is required
      skip_inactive (boolean, default True): if True, connections are not included where there is an inactive cell
         above or below the pinchout; if False, such connections are included
      feature_name (string, default 'pinchout'): the name to use as citation title in the feature and interpretation
      tolerance (float, default 0.001): the minimum vertical distance below which a K gap is deemed to be zero
         thickness; units are implicitly the z units of the coordinate reference system used by grid

   notes:
      this function does not write to hdf5, nor create xml for the new grid connection set;
      however, it does create one feature and a corresponding interpretation and creates xml for those;
      note that the entries in the connection set will be for logically K-neighbouring pairs of cells – such pairs
      are omitted from the standard transmissibilities due to the presence of the K gap layer
   """

   assert grid is not None
   if not grid.k_gaps: return None

   p = grid.points_ref(masked = False)
   dead = grid.extract_inactive_mask() if skip_inactive else None
   flip_z = (grid.k_direction_is_down != rqet.find_tag_bool(grid.crs_root, 'ZIncreasingDownward'))

   cip_list = []  # cell index pair list

   for k in range(grid.nk - 1):
      if grid.k_gap_after_array[k]:
         k_gap_pillar_z = p[grid.k_raw_index_array[k + 1]][..., 2] - p[grid.k_raw_index_array[k] + 1][..., 2]
         if grid.has_split_coordinate_lines:
            pfc = grid.create_column_pillar_mapping()  # pillars for column
            k_gap_z = 0.25 * np.sum(k_gap_pillar_z[pfc], axis = (2, 3))  # resulting shape (nj, ni)
         else:
            k_gap_z = 0.25 * (k_gap_pillar_z[:-1, :-1] +
                              k_gap_pillar_z[:-1, 1:] +
                              k_gap_pillar_z[1:, :-1] +
                              k_gap_pillar_z[1:, 1:])  # shape (nj, ni)
      if flip_z: k_gap_z = -k_gap_z
      layer_mask = np.logical_and(np.logical_not(np.isnan(k_gap_z)), k_gap_z < tolerance)
      if skip_inactive:
         layer_mask = np.logical_and(layer_mask, np.logical_not(np.logical_or(dead[k], dead[k + 1])))
      # layer mask now boolean array of shape (nj, ni) set True where connection needed
      ji_list = np.stack(np.where(layer_mask)).T  # numpy array being list of [j, i] pairs
      for (j, i) in ji_list:
         cip_list.append((grid.natural_cell_index((k, j, i)), grid.natural_cell_index((k + 1, j, i))))

   log.debug(f'{len(cip_list)} k gap connections found')

   kgcs = _make_k_gcs_from_cip_list(grid, cip_list, feature_name)

   return kgcs



def add_connection_set_and_tmults(model, fault_incl, tmult_dict):
   """Add a grid connection set to a resqml model, based on a fault include file and a dictionary of fault:tmult pairs.

   Grid connection set added to resqml model, with extra_metadata on the fault interpretation containing the MULTFL values

   Args:
      model - resqml model object
      fault_incl - fullpath to fault include file or list of fullpaths to fault include files
      tmult_dict - dictionary of fault name: tmult pairs (must align with faults in include file)
   Returns:
      grid connection set uuid
   """
   for fault in tmult_dict.keys():
       log.debug(f"Working on {fault}")
       tbf = rqo.TectonicBoundaryFeature(parent_model=model, feature_name=fault, kind='fault')
       tbf.create_xml()
       fint = rqo.FaultInterpretation(parent_model=model, tectonic_boundary_feature=tbf, is_normal=True)
       fint.extra_metadata = {"Transmissibility multiplier":tmult_dict[fault]}
       fint.create_xml()
       model.store_epc()
       log.info(f"{fault} added to RESQML model")
   log.info("Creating grid connection set")
   if isinstance(fault_incl, list):
        if len(fault_incl) > 1:
            # Making a concatenated version of the faultincl files
            # TODO: perhaps a better/more unique name and location could be used in future?
            temp_faults = os.path.join(os.path.dirname(model.epc_file), 'faults_combined_temp.txt')
            with open(temp_faults, 'w') as outfile:
                log.debug("combining multiple include files into one")
                for fname in fault_incl:
                    with open(fname) as infile:
                        for line in infile:
                            outfile.write(line)
        else: temp_faults = fault_incl[0]
   else: temp_faults = fault_incl
   gcs = GridConnectionSet(parent_model = model, ascii_load_format = 'nexus', ascii_file = temp_faults)
   gcs.write_hdf5(model.h5_file_name())
   gcs.create_xml(model.h5_uuid())
   model.store_epc()
   if os.path.exists(os.path.join(os.path.dirname(model.epc_file), 'faults_combined_temp.txt')):
      os.remove(os.path.join(os.path.dirname(model.epc_file), 'faults_combined_temp.txt'))
   log.info("Grid connection set added")

   return gcs.uuid



# fault face table pandas dataframe functions
# these functions are for processing dataframes that have been read from (or to be written to) simulator ascii files

def zero_base_cell_indices_in_faces_df(faces, reverse = False):
   """Decrements all the cell indices in the fault face dataframe, in situ (or increments if reverse is True)."""

   if reverse: offset = 1
   else: offset = -1
   for col in ['i1', 'i2', 'j1', 'j2', 'k1', 'k2']:
      temp = faces[col] + offset
      faces[col] = temp



def standardize_face_indicator_in_faces_df(faces):
   """Sets face indicators to uppercase I, J or K, always with + or - following direction, in situ."""

   # todo: convert XYZ into IJK respectively?
   temp = faces['face'].copy().str.upper()
   temp[faces['face'].str.len() == 1] = faces['face'][faces['face'].str.len() == 1] + '+'
   for u in temp.unique():
      s = str(u)
      assert len(s) == 2, 'incorrect length to face string in fault dataframe: ' + s
      assert s[0] in 'IJK', 'unknown direction (axis) character in fault dataframe: ' + s
      assert s[1] in '+-', 'unknown face polarity character in fault dataframe: ' + s
   faces['face'] = temp



def remove_external_faces_from_faces_df(faces, extent_kji, remove_all_k_faces = False):
   """Returns a subset of the rows of faces dataframe, excluding rows on external faces."""

   # NB: assumes cell indices have been converted to zero based
   # NB: ignores grid column, ie. assumes extent_kji is applicable to all rows
   # NB: assumes single layer of cells is specified in the direction of the face
   filtered = []
   max_k0 = extent_kji[0] - 1
   max_j0 = extent_kji[1] - 1
   max_i0 = extent_kji[2] - 1
   for i in range(len(faces)):
      entry = faces.iloc[i]
      f = entry['face']
      if ((entry['i1'] <= 0 and f == 'I-') or
          (entry['j1'] <= 0 and f == 'J-') or
          (entry['k1'] <= 0 and f == 'K-') or
          (entry['i2'] >= max_i0 and f == 'I+') or
          (entry['j2'] >= max_j0 and f == 'J+') or
          (entry['k2'] >= max_k0 and f == 'K+')): continue
      if remove_all_k_faces and f[0] == 'K': continue
      filtered.append(i)
   return faces.loc[filtered]



def _make_k_gcs_from_cip_list(grid, cip_list, feature_name):
   # cip (cell index pair) list contains pairs of natural cell indices for which k connection is required
   # first of pair is layer above (lower k to be precise), second is below (higher k)
   # called by pinchout_connection_set() and k_gap_connection_set() functions

   count = len(cip_list)

   if count == 0: return None

   pcs = GridConnectionSet(grid.model)
   pcs.grid_list = [grid]
   pcs.count = count
   pcs.grid_index_pairs = np.zeros((count, 2), dtype = int)
   pcs.cell_index_pairs = np.array(cip_list, dtype = int)
   pcs.face_index_pairs = np.zeros((count, 2), dtype = int)  # initialize to top faces
   pcs.face_index_pairs[:, 0] = 1  # bottom face of cells above pinchout

   pcs.feature_indices = np.zeros(count, dtype = int)  # could create seperate features by layer above or below?
   gbf = rqo.GeneticBoundaryFeature(grid.model, kind = 'horizon', feature_name = feature_name)
   gbf_root = gbf.create_xml()
   fi = rqo.HorizonInterpretation(grid.model, genetic_boundary_feature = gbf)
   fi_root = fi.create_xml(gbf_root, title_suffix = None)
   fi_uuid = rqet.uuid_for_part_root(fi_root)

   pcs.feature_list = [('obj_HorizonInterpretation', fi_uuid, str(feature_name))]

   return pcs



# test module for resqpy.olio.transmission.py

import pytest
import os
import numpy as np

import resqpy.model as rq
import resqpy.grid as grr
import resqpy.olio.transmission as rqtr
import resqpy.olio.vector_utilities as vec
import resqpy.olio.uuid as bu

def test_regular_grid_half_cell_transmission(tmp_path):

   def try_one_half_t_regular(model, extent_kji = (2, 2, 2), dxyz = (1.0, 1.0, 1.0),
                              perm_kji = (1.0, 1.0, 1.0), ntg = 1.0, darcy_constant = 1.0,
                              rotate = None, dip = None):
       ones = np.ones(extent_kji)
       grid = grr.RegularGrid(model, extent_kji = extent_kji, dxyz = dxyz)
       if dip is not None:  # dip positive x axis downwards
          r_matrix = vec.rotation_matrix_3d_axial(1, dip)
          p = grid.points_ref(masked = False)
          p[:] = vec.rotate_array(r_matrix, p)
       if rotate is not None:  # rotate anticlockwise in xy plane (viewed from above)
          r_matrix = vec.rotation_matrix_3d_axial(2, rotate)
          p = grid.points_ref(masked = False)
          p[:] = vec.rotate_array(r_matrix, p)
       half_t = rqtr.half_cell_t(grid,
                                 perm_k = perm_kji[0] * ones,
                                 perm_j = perm_kji[1] * ones,
                                 perm_i = perm_kji[2] * ones,
                                 ntg = ntg * ones,
                                 darcy_constant = darcy_constant)
       expected = 2.0 * darcy_constant * np.array((perm_kji[0] * dxyz[0] * dxyz[1] / dxyz[2],
                                                   ntg * perm_kji[1] * dxyz[0] * dxyz[2] / dxyz[1],
                                                   ntg * perm_kji[2] * dxyz[1] * dxyz[2] / dxyz[0]))
       assert np.all(np.isclose(half_t, expected.reshape(1, 1, 1, 3)))

   temp_epc = str(os.path.join(tmp_path, f"{bu.new_uuid()}.epc"))
   model = rq.Model(temp_epc, new_epc = True, create_basics = True)

   try_one_half_t_regular(model)
   try_one_half_t_regular(model, extent_kji = (3, 4, 5))
   try_one_half_t_regular(model, dxyz = (127.53, 21.05, 12.6452))
   try_one_half_t_regular(model, perm_kji = (123.23, 512.4, 314.7))
   try_one_half_t_regular(model, ntg = 0.7)
   try_one_half_t_regular(model, darcy_constant = 0.001127)
   try_one_half_t_regular(model, extent_kji = (5, 4, 3),
                                 dxyz = (84.23, 77.31, 15.823),
                                 perm_kji = (0.6732, 298.14, 384.2),
                                 ntg = 0.32,
                                 darcy_constant = 0.008527)
   try_one_half_t_regular(model, rotate = 67.8)

   # should not have written anything, but try clean-up just in case
   try:
      os.remove(temp_epc)
   except:
      pass
   try:
      os.remove(temp_epc[-4] + '.h5')
   except:
      pass



import os

import pytest
import numpy as np

import resqpy.model as rq
import resqpy.grid as grr
import resqpy.property as rqp
import resqpy.derived_model as rqdm
import resqpy.olio.weights_and_measures as bwam


# ---- Test PropertyCollection ---

# TODO


# ---- Test Property ---

def test_property(tmp_path):
   epc = os.path.join(tmp_path, 'test_prop.epc')
   model = rq.new_model(epc)
   grid = grr.RegularGrid(model, extent_kji = (2, 3, 4))
   grid.write_hdf5()
   grid.create_xml()
   a1 = np.random.random(grid.extent_kji)
   p1 = rqp.Property.from_array(model, a1, source_info = 'random', keyword = 'NETGRS', support_uuid = grid.uuid,
                                property_kind = 'net to gross ratio', indexable_element = 'cells', uom = 'm3/m3')
   pk = rqp.PropertyKind(model, title = 'facies', parent_property_kind = 'discrete')
   pk.create_xml()
   facies_dict = {0: 'background'}
   for i in range(1, 10): facies_dict[i] = 'facies ' + str(i)
   sl = rqp.StringLookup(model, int_to_str_dict = facies_dict, title = 'facies table')
   sl.create_xml()
   shape2 = tuple(list(grid.extent_kji) + [2])
   a2 = (np.random.random(shape2) * 10.0).astype(int)
   p2 = rqp.Property.from_array(model, a2, source_info = 'random', keyword = 'FACIES', support_uuid = grid.uuid,
                                local_property_kind_uuid = pk.uuid, indexable_element = 'cells', discrete = True,
                                null_value = 0, count = 2, string_lookup_uuid = sl.uuid)
   model.store_epc()
   model = rq.Model(epc)
   ntg_uuid = model.uuid(obj_type = p1.resqml_type, title = 'NETGRS')
   assert ntg_uuid is not None
   p1p = rqp.Property(model, uuid = ntg_uuid)
   assert np.all(p1p.array_ref() == p1.array_ref())
   assert p1p.uom() == 'm3/m3'
   facies_uuid = model.uuid(obj_type = p2.resqml_type, title = 'FACIES')
   assert facies_uuid is not None
   p2p = rqp.Property(model, uuid = facies_uuid)
   assert np.all(p2p.array_ref() == p2.array_ref())
   assert p2p.null_value() is not None and p2p.null_value() == 0
   grid = model.grid()
   assert grid.property_collection.number_of_parts() == 5  # two created here, plus 3 regular grid cell lengths properties


def test_create_Property_from_singleton_collection(tmp_model):

   # Arrange
   grid = grr.RegularGrid(tmp_model, extent_kji = (2, 3, 4))
   grid.write_hdf5()
   grid.create_xml(add_cell_length_properties = True)
   collection = rqp.selective_version_of_collection(grid.property_collection, property_kind = 'cell length',
                                                    facet_type = 'direction', facet='J')
   assert collection.number_of_parts() == 1
   # Check property can be created
   prop = rqp.Property.from_singleton_collection(collection)
   assert np.all(np.isclose(prop.array_ref(), 1.0))


# ---- Test property kind parsing ---


def test_property_kinds():
   prop_kinds = bwam.valid_property_kinds()
   assert "angle per time" in prop_kinds
   assert "foobar" not in prop_kinds


# ---- Test infer_property_kind ---

@pytest.mark.parametrize("input_name, input_unit, kind, facet_type, facet", [
   ('gamma ray API unit', 'gAPI', 'gamma ray API unit', None, None),  # Just a placeholder for now
   ('', '', 'Unknown', None, None),  # What should default return?
])
def test_infer_property_kind(input_name, input_unit, kind, facet_type, facet):
   kind_, facet_type_, facet_ = rqp.infer_property_kind(input_name, input_unit)
   assert kind == kind_
   assert facet_type == facet_type_
   assert facet == facet_


# ---- Test bespoke property kind reuse ---

def test_bespoke_property_kind():
   model = rq.Model(create_basics = True)
   em = {'something': 'important', 'and_another_thing': 42}
   pk1 = rqp.PropertyKind(model, title = 'my kind of property', extra_metadata = em)
   pk1.create_xml()
   assert len(model.uuids(obj_type = 'PropertyKind')) == 1
   pk2 = rqp.PropertyKind(model, title = 'my kind of property', extra_metadata = em)
   pk2.create_xml()
   assert len(model.uuids(obj_type = 'PropertyKind')) == 1
   pk3 = rqp.PropertyKind(model, title = 'your kind of property', extra_metadata = em)
   pk3.create_xml()
   assert len(model.uuids(obj_type = 'PropertyKind')) == 2
   pk4 = rqp.PropertyKind(model, title = 'my kind of property')
   pk4.create_xml()
   assert len(model.uuids(obj_type = 'PropertyKind')) == 3
   pk5 = rqp.PropertyKind(model, title = 'my kind of property', parent_property_kind = 'discrete', extra_metadata = em)
   pk5.create_xml()
   assert len(model.uuids(obj_type = 'PropertyKind')) == 4
   pk6 = rqp.PropertyKind(model, title = 'my kind of property', parent_property_kind = 'continuous')
   pk6.create_xml()
   assert len(model.uuids(obj_type = 'PropertyKind')) == 4
   assert pk6 == pk4


# ---- Test string lookup ---

def test_string_lookup():
   model = rq.Model(create_basics = True)
   d = {1: 'peaty',
        2: 'muddy',
        3: 'sandy',
        4: 'granite'}
   sl = rqp.StringLookup(model, int_to_str_dict = d, title = 'stargazing')
   sl.create_xml()
   assert sl.length() == 5 if sl.stored_as_list else 4
   assert sl.get_string(3) == 'sandy'
   assert sl.get_index_for_string('muddy') == 2
   d2 = {0: 'nothing',
         27: 'something',
         173: 'amazing',
         1072: 'brilliant'}
   sl2 = rqp.StringLookup(model, int_to_str_dict = d2, title = 'head in the clouds')
   assert not sl2.stored_as_list
   assert sl2.length() == 4
   assert sl2.get_string(1072) == 'brilliant'
   assert sl2.get_string(555) is None
   assert sl2.get_index_for_string('amazing') == 173
   sl2.create_xml()
   assert set(model.titles(obj_type = 'StringTableLookup')) == set(['stargazing', 'head in the clouds'])
   assert sl != sl2



def test_property_extra_metadata(tmp_path):
   # set up test model with a grid
   epc = os.path.join(tmp_path, 'em_test.epc')
   model = rq.new_model(epc)
   grid = grr.RegularGrid(model, extent_kji = (2,3,4))
   grid.write_hdf5()
   grid.create_xml()
   model.store_epc()
   # add a grid property with some extra metadata
   a = np.arange(24).astype(float).reshape((2, 3, 4))
   em = {'important': 'something', 'also': 'nothing'}
   uuid = rqdm.add_one_grid_property_array(epc, a, 'length', title = 'nonsense', uom = 'm', extra_metadata = em)
   # re-open the model and check that extra metadata has been preserved
   model = rq.Model(epc)
   p = rqp.Property(model, uuid = uuid)
   for item in em.items():
       assert item in p.extra_metadata.items()



# test hexahedral volume calculations

import math as maths
import numpy as np
from numpy.random import random
from numpy.testing import assert_array_almost_equal

import resqpy.olio.volume as vol

def unit_cube():
   cp = np.zeros((2, 2, 2, 3))  # pagoda style corner point array for a single cell
   # set to unit cube with xyz aligned with ijk respectively
   cp[:, :, 1, 0] = 1.0
   cp[:, 1, :, 1] = 1.0
   cp[1, :, :, 2] = 1.0
   return cp


def test_one_cell_volume():
   cp = unit_cube().copy()
   assert maths.isclose(vol.tetra_cell_volume(cp, off_hand = False), 1.0), 'unit cube cell volume calculation fails'
   # stretch in x (I)
   cp[:, :, 1, 0] *= 2.5
   assert maths.isclose(vol.tetra_cell_volume(cp, off_hand = False), 2.5), 'square prism cell volume calculation fails'
   # stretch in y & z (J & K)
   cp[:, 1, :, 1] *= 2.0
   cp[1, :, :, 2] *= 3.0
   assert maths.isclose(vol.tetra_cell_volume(cp, off_hand = False), 15.0), 'regular cuboid cell volume calculation fails'
   # apply a shear in x & y (should not change volume)
   cp[1, :, :, :2] += 0.5
   assert maths.isclose(vol.tetra_cell_volume(cp, off_hand = False), 15.0), 'parallelepiped cell volume calculation fails'
   # translate into negative coordinate realm
   cp[:] -= 3.827
   assert maths.isclose(vol.tetra_cell_volume(cp, off_hand = False), 15.0), 'translated cell volume calculation fails '
   # invert one axis
   cp[:, :, :, 1] *= -1.0
   assert maths.isclose(vol.tetra_cell_volume(cp, off_hand = True), 15.0), 'off hand True cell volume calculation fails'
   # collapse to plane
   cp[:, :, :, 2] = 0.0
   assert maths.isclose(vol.tetra_cell_volume(cp, off_hand = False), 0.0), 'planar cell volume calculation fails'
   assert maths.isclose(vol.tetra_cell_volume(cp, off_hand = True), 0.0), 'off hand True planar cell volume calculation fails'
   # test wedge
   cp = unit_cube().copy()
   cp[1, 1, :, 2] = 0.0
   assert maths.isclose(vol.tetra_cell_volume(cp, off_hand = False), 0.5), 'triangular prism volume calculation fails'


def test_array_volume():
   # note: the multiple cells generated here are all overlapping, which would not usually be the case
   cp = np.zeros((5, 10, 15, 2, 2, 2, 3))  # pagoda style corner point array
   cp[:] = unit_cube().reshape((1, 1, 1, 2, 2, 2, 3)) * 2.0
   cp[:] += random(cp.size).reshape(cp.shape) - 0.5
   # move whole cells by random translations
   translate = (random(cp.size // 8) * 123.2356).reshape((cp.shape[0], cp.shape[1], cp.shape[2], 1, 1, 1, 3))
   cp[:] += translate
   # compute volumes using two different routines in the volume module and make sure they give the same answer
   v1 = vol.tetra_volumes_slow(cp, off_hand = False)
   v2 = vol.tetra_volumes(cp, off_hand = False)
   # following is a stringent test requiring that exactly the same mathematical operations have been performed
   assert_array_almost_equal(v1, v2)

   # the random changes to corner points should have left the volumes within a certain range
   assert np.all(v1 >= 0.0), 'negative volume(s) returned by array function'
   assert np.all(v1 <= 27.0), 'exaggerated volume(s) returned by array function'
   # translate again and check that volumes are unchanged
   translate = (random(cp.size // 8) * 3.23478).reshape((cp.shape[0], cp.shape[1], cp.shape[2], 1, 1, 1, 3))
   cp[:] -= translate
   v2 = vol.tetra_volumes(cp, off_hand = False)
   assert_array_almost_equal(v1, v2)

   # test handedness inversion
   cp[:, :, :, :, :, :, 0] *= -1.0
   v1 = vol.tetra_volumes_slow(cp, off_hand = True)
   v2 = vol.tetra_volumes(cp, off_hand = True)
   assert_array_almost_equal(v1, v2)
   assert np.all(v1 >= 0.0), 'negative volume(s) returned by array function'
   assert np.all(v1 <= 27.0), 'exaggerated volume(s) returned by array function'



"""point_inclusion.py: functions to test whether a point is within a polygon; also line intersection with planes."""

version = '29th April 2021'

import logging
log = logging.getLogger(__name__)
log.debug('point_inclusion.py version ' + version)

import os
import math as maths
import numpy as np

import resqpy.olio.simple_lines as sl


def pip_cn(p, poly):
   """2D point inclusion: returns True if point is inside polygon (uses crossing number algorithm)"""

   # p should be a tuple-like object with first two elements x, y
   # poly should be a numpy array with first axis being the different vertices
   #    and the second starting x, y

   crossings = 0
   vertex_count = poly.shape[0]

   for edge in range(vertex_count):
      v1 = poly[edge]
      if edge == vertex_count - 1: v2 = poly[0]
      else: v2 = poly[edge + 1]
      if ((v1[0] > p[0] or v2[0] > p[0]) and
          ((v1[1] <= p[1] and v2[1] > p[1]) or (v1[1] > p[1] and v2[1] <= p[1]))):
         if v1[0] > p[0] and v2[0] > p[0]: crossings += 1
         elif p[0] < v1[0] + (v2[0] - v1[0]) * (p[1] - v1[1]) / (v2[1] - v1[1]): crossings += 1

   return bool(crossings & 1)


def pip_wn(p, poly):
   """2D point inclusion: returns True if point is inside polygon (uses winding number algorithm)"""

   winding = 0
   vertex_count = poly.shape[0]

   for edge in range(vertex_count):
      v1 = poly[edge]
      if edge == vertex_count - 1: v2 = poly[0]
      else: v2 = poly[edge + 1]
      if v1[0] <= p[0] and v2[0] <= p[0]: continue
      if v1[1] <= p[1] and v2[1] > p[1]:
         if v1[0] > p[0] and v2[0] > p[0]: winding += 1
         elif p[0] < v1[0] + (v2[0] - v1[0]) * (p[1] - v1[1]) / (v2[1] - v1[1]): winding += 1
      elif v1[1] > p[1] and v2[1] <= p[1]:
         if v1[0] > p[0] and v2[0] > p[0]: winding -= 1
         elif p[0] < v1[0] + (v2[0] - v1[0]) * (p[1] - v1[1]) / (v2[1] - v1[1]): winding -= 1

   return bool(winding)


def pip_array_cn(p_a, poly):
   """array of 2D points inclusion: returns bool array True where point is inside polygon (uses crossing number algorithm)"""

   # p_array should be a numpy array of 2 or more axes; the final axis has extent at least 2, being x, y, ...
   # returned boolean array has shape of p_array less the final axis

   log.debug('type(poly): ' + str(type(poly)))
   elements = np.prod(list(p_a.shape)[:-1], dtype = int)
   log.debug('elements: ' + str(elements))
   p = p_a.reshape((elements, -1))
   log.debug('p.shape: ' + str(p.shape))
   crossings = np.zeros((elements,), dtype = int)
   vertex_count = poly.shape[0]
   log.debug('vertex_count: ' + str(vertex_count))

   np_err_dict = np.geterr()
   np.seterr(divide = 'ignore', invalid = 'ignore')
   for edge in range(vertex_count):
      v1 = poly[edge]
      if edge == vertex_count - 1: v2 = poly[0]
      else: v2 = poly[edge + 1]
      crossings += np.where(np.logical_and(np.logical_and(np.logical_or(v1[0] > p[:, 0], v2[0] > p[:, 0]),
                                                          np.logical_or(np.logical_and(v1[1] <= p[:, 1], v2[1] > p[:, 1]),
                                                                        np.logical_and(v1[1] > p[:, 1], v2[1] <= p[:, 1]))),
                                           np.logical_or(np.logical_and(v1[0] > p[:, 0], v2[0] > p[:, 0]),
                                                         (p[:, 0] < (v1[0] + (v2[0] - v1[0]) * (p[:, 1] - v1[1]) / (v2[1] - v1[1]))))),
                            1, 0)
   if 'divide' in np_err_dict: np.seterr(divide = np_err_dict['divide'])
   if 'invalid' in np_err_dict: np.seterr(invalid = np_err_dict['invalid'])

   return np.array(np.bitwise_and(crossings, 1), dtype = bool).reshape(list(p_a.shape)[:-1])


def points_in_polygon(x, y, polygon_file, poly_unit_multiplier = None):
   """Takes a pair of numpy arrays x, y defining points to be tested against polygon specified in polygon_file."""

   assert x.shape == y.shape, 'x and y arrays have differing shapes or are not numpy arrays'
   assert polygon_file and os.path.exists(polygon_file), 'polygon file is missing'
   if poly_unit_multiplier is not None: assert poly_unit_multiplier != 0.0, 'zero multiplier for polygon units not allowed'

   try:
      polygon_list = sl.read_lines(polygon_file)
      assert len(polygon_list) > 0, 'unable to read polygon from file ' + polygon_file
      assert len(polygon_list) == 1, 'more than one polygon in file ' + polygon_file
      polygon = polygon_list[0]
      if poly_unit_multiplier is not None: polygon[:] *= poly_unit_multiplier
      points = np.stack((x, y), axis = -1)
      return pip_array_cn(points, polygon)
   except:
      log.exception('failed to determine points in polygon')

   return None


def scan(origin, ncol, nrow, dx, dy, poly):
   """Scans lines of pixels returning 2D boolean array of points within convex polygon."""

   vertex_count = poly.shape[0]
   inside = np.zeros((nrow, ncol), dtype = bool)
   ix = np.empty((2))

   for row in range(nrow):
      ix[:] = 0.0
      y = origin[1]  +  dy * row
      ic = 0
      for edge in range(vertex_count):
         v1 = poly[edge]
         if edge == vertex_count - 1: v2 = poly[0]
         else: v2 = poly[edge + 1]
         if (v1[1] > y and v2[1] > y) or (v1[1] <= y and v2[1] <= y): continue
         ix[ic] = v1[0]  +  (v2[0] - v1[0]) * (y - v1[1]) / (v2[1] - v1[1])
         ic += 1
         if ic == 2: break
      if ic < 2: continue
      if ix[1] < ix[0]: sx, ex = ix[1], ix[0]
      else: sx, ex = ix[0], ix[1]
      scol = maths.ceil((sx - origin[0]) / dx)
      ecol = int((ex - origin[0]) / dx)
      inside[row, scol:ecol+1] = True

   return inside



#  write_faults.py

version = '29th April 2021'

# Nexus is a registered trademark of the Halliburton Company

import logging
log = logging.getLogger(__name__)
log.debug('write_faults.py version ' + version)

import pandas as pd

import resqpy.olio.trademark as tm

#  both functions expect df to be a pandas dataframe with columns:
#   name, i1, i2, j1, j2, k1, k2, face
# where name is the fault/baffle name
# i,j,k indices are in simulator protocol (one based)
# face is either I+ or J+ for now


def write_faults_nexus(filename, df, grid_name = 'ROOT'):
   """Creates a Nexus include file holding MULT keywords with FNAME and face data, from pandas dataframe."""

   def write_header_lines(fp, T, grid_name, fault_name):
      fp.write('\nMULT\t' + T + '\tALL\tPLUS\tMULT\n')
      fp.write('\tGRID\t' + grid_name + '\n')
      fp.write('\tFNAME\t' + fault_name + '\n')
      if len(fault_name) > 256:
         log.warning('exported fault name longer than Nexus limit of 256 characters: ' + fault_name)
         tm.log_nexus_tm('warning')


   def write_rows(fp, df):
      for row in range(len(df)):
         fp.write('\t{0:1d}\t{1:1d}\t{2:1d}\t{3:1d}\t{4:1d}\t{5:1d}\t1.0\n'.format(
            df.iloc[row, 1], df.iloc[row, 2], df.iloc[row, 3], df.iloc[row, 4], df.iloc[row, 5], df.iloc[row, 6]))

   log.info('writing FNAME data in Nexus format to file: ' + filename)
   tm.log_nexus_tm('info')
   assert df is not None and len(df) > 0, 'no data in faults dataframe'

   with open(filename, 'w') as fp:
      fp.write('! Exported from Bifrost resqml_baffle notebook\n\n')
      fault_names = pd.unique(df.name)
      for fault_name in fault_names:
         fdf = df[df.name == fault_name]
         fdfi = fdf[fdf.face == 'I+']
         fdfj = fdf[fdf.face == 'J+']
         if len(fdfi):
            write_header_lines(fp, 'TX', grid_name, fault_name)
            write_rows(fp, fdfi)
         if len(fdfj):
            write_header_lines(fp, 'TY', grid_name, fault_name)
            write_rows(fp, fdfj)



# load_data module
# contains functions to load data from various formats of file

version = '29th April 2021'

# Nexus is a registered trademark of the Halliburton Company

import logging
log = logging.getLogger(__name__)
log.debug('load_data.py version %s', version)

# defs:
#    def file_exists(file_name):
#    def load_corp_array_from_file(file_name, extent_kji = None, comment_char = '!',
#                                  data_free_of_comments = False, use_binary = False,
#                                  eight_mode = False):
#    def load_array_from_file(file_name, extent, data_type = 'real', comment_char = '!',
#                             data_free_of_comments = False, use_binary = False):
#    def load_array_from_ascii_file(file_name, extent, data_type='real', comment_char = '!', data_free_of_comments = False):
#    def load_fault_mask(file_name, direction, fault_mask, use_binary = False):

import os
import numpy as np

import resqpy.olio.utility as pu
import resqpy.olio.box_utilities as box
import resqpy.olio.keyword_files as kf
import resqpy.olio.write_data as wd
import resqpy.olio.grid_functions as gf
import resqpy.olio.ab_toolbox as abt


def file_exists(file_name, must_be_more_recent_than_file = None):
   """Returns True if the file exists (and is more recent than other file, if given)."""

   if file_name is None or len(file_name) == 0: return False
   existe = os.path.exists(file_name)
   if not existe: return False
   if not must_be_more_recent_than_file or file_name == must_be_more_recent_than_file or  \
      not os.path.exists(must_be_more_recent_than_file): return True
   return os.path.getmtime(file_name) > os.path.getmtime(must_be_more_recent_than_file)


######################################################################################################
# load_corp_array_from_file():
# function to load a nexus corp array of data from a named file
# if the grid extent is not known, the file must be free of comments
# NB: extent_kji here is extent of grid, rather than that of the corp array
# returns a pagoda style 7D array, resequenced

def load_corp_array_from_file(file_name, extent_kji = None, corp_bin = False, swap_bytes = True,
                              max_lines_for_keyword = 100, comment_char = None,
                              data_free_of_comments = False, use_binary = False,
                              eight_mode = False, use_numbers_only = None):
   """Loads a nexus corner point (CORP) array from a file, returns a 7D numpy array in pagoda ordering.

      arguments:
         file_name: The name of an ascii file holding the CORP data (no other keywords with numeric data
                  should be in the file); write access to the directory is likely to be needed if
                  use_binary is True
         extent_kji: The extent of the grid as a list or a 3 element numpy array, in the order [NK, NJ, NI].
                  If extent_kji is None, the extent is figured out from the data.  It must be
                  given for 1D or 2D models
         corp_bin (boolean, default False): if True, input file is in bespoke corp binary format, otherwise ascii
         swap_bytes (boolean, default True): if True, byte ordering of corp bin data is reversed; only relevant
                  if corp_bin is True
         max_lines_for_keyword: the maximum number of lines to search for CORP keyword; set to zero if file is
                  known to be data only
         comment_char: A single character string which is interpreted as introducing a comment
         data_free_of_comments: If True, once the numeric data is encountered, it is assumed that there are no
                  further comments (allowing a faster load)
         use_binary: If True, a more recent file containing a pure binary copy of the data is looked for first,
                  in the same directory; if found, the data is loaded directly from that file; if not found, the
                  binary file is created after the ascii has been loaded (ready for next time)
         eight_mode: If True, the data is assumed to be in CORP EIGHT ordering; otherwise the normal ordering
                  (The code does not look for keywords.); this is not automatically determined from any keyword
                  in the file
         use_numbers_only: no longer in use, ignored

      returns:
         A numpy array containing the CORP data in 7D pagoda protocol ordering. The extent of the grid, and hence
         shape of the array is determined from the corner point data unless extent_kji has been specified.
   """

#   assert(extent_kji is not None or data_free_of_comments)   # no longer a requirement

   if extent_kji is None:
      extent = None
   else:
      extent = [extent_kji[0] * extent_kji[1] * extent_kji[2], 8, 3]

   if corp_bin:
      bin_file_size = os.path.getsize(file_name)
      bin_cell_count, remainder = divmod(bin_file_size, 26 * 4)   # corp bin format has records of 24 + 2 (head, tail) 32 bit words
      if remainder:
         log.error('corp binary file ' + str(file_name) + ' is not a whole number of records')
         return None
      dt = np.dtype('float32')
      if swap_bytes: dt = dt.newbyteorder()
      cp_array = np.fromfile(file_name, dtype = dt, count = 26 * bin_cell_count).reshape((-1, 26))[:, 1:25]
      if extent_kji is not None:
         if bin_cell_count != (extent_kji[0] * extent_kji[1] * extent_kji[2]):
            log.error('corp binary file ' + str(file_name) + ' contains data for ' + str(bin_cell_count) +
                      ' cells; extent requires ' + str(extent_kji[0] * extent_kji[1] * extent_kji[2]))

   else:
      cp_array = load_array_from_file(file_name, extent = extent, data_type = 'real',
                                      keyword = 'CORP', max_lines_for_keyword = max_lines_for_keyword,
                                      comment_char = comment_char,
                                      data_free_of_comments = data_free_of_comments,
                                      use_binary = use_binary)

   cell_count, remainder = divmod(cp_array.size, 24)
   if remainder:
      log.error('file ' + file_name + ' contains ' + str(cp_array.size) + ' data, which is not a multiple of 24')
      return None

   cp_array = cp_array.reshape(1, 1, cell_count, 2, 2, 2, 3)  # pagoda 7D, temporarily with all cells on I axis
   log.debug('resequencing corner point data')
   gf.resequence_nexus_corp(cp_array, eight_mode = eight_mode)  # switches IP points where JP = 1 (unless eight_mode)

   if extent_kji is None:
      log.info('determining grid extent from corner points')
      extent_kji = gf.determine_corp_extent(cp_array)
      if extent_kji is None:
         log.error('failed to determine extent of grid from corner points')
         return cp_array

   return cp_array.reshape(extent_kji[0], extent_kji[1], extent_kji[2], 2, 2, 2, 3)


######################################################################################################
# load_array_from_file():
# function to load an array of data from a named file
# if the extent is not known (None), the file must be free of comments
# a new numpy array is returned
# binary data file creation & reuse is supported

def load_array_from_file(file_name, extent = None, data_type = 'real', keyword = None,
                         max_lines_for_keyword = None, comment_char = None,
                         data_free_of_comments = False, use_binary = False, use_numbers_only = None):
   """Load an array from an ascii (or pure binary) file.

      Arguments are similar to those for load_corp_array_from_file().
   """

   if not use_binary:
      return load_array_from_ascii_file(file_name, extent = extent, data_type = data_type,
                                        keyword = keyword, max_lines_for_keyword = max_lines_for_keyword,
                                        comment_char = comment_char,
                                        data_free_of_comments = data_free_of_comments)

   (extension, ab_type) = abt.binary_file_extension_and_np_type_for_data_type(data_type)

   if extent is None:
       cell_count = -1   # np.fromfile interprets this as 'read everything'
       log.debug('Loading unknown number of array data elements from file ' + file_name)
   else:
       cell_count = pu.cell_count_from_extent(extent)
       log.debug('Loading %1d array data elements from file %s', cell_count, file_name)

   ascii_file_name = file_name
   binary_file_name = file_name
   if len(binary_file_name) < 4 or binary_file_name[-3:] != extension: binary_file_name += extension
   else: ascii_file_name = ascii_file_name[:-3]   # strip off '.db' or similar

   try:   # tentatively try to read data from an existing binary file, if present
      if file_exists(binary_file_name, must_be_more_recent_than_file = ascii_file_name):
         with open(binary_file_name, 'rb') as binary_file_in:
            result = np.fromfile(binary_file_in, dtype = ab_type, count = cell_count)
            if extent is not None: result = result.reshape(extent)
            # check that end of file has been reached, ie. not too much data in file
            try: # expected to return null
               c = binary_file_in.read(1)
               if len(c): log.warning('binary file contains more data than expected: ' + binary_file_name)
            except:
               pass
            log.info('Data loaded from binary file %s', binary_file_name)
            return result
   except:
      pass

   # read from ascii file
   result = load_array_from_ascii_file(file_name, extent = extent, data_type = data_type,
                                       keyword = keyword, max_lines_for_keyword = max_lines_for_keyword,
                                       comment_char = comment_char,
                                       data_free_of_comments = data_free_of_comments)

   # create a binary file (to be used next time)
   try:
      wd.write_pure_binary_data(binary_file_name, result)
   except:
      log.warn('Failed to write data to binary file %s', binary_file_name)
      # todo: could delete the binary file in case a corrupt file is left for use next time

   return result

# end of load_array_from_file() def
######################################################################################################


######################################################################################################
# load_array_from_ascii_file():
# function to load an array of data from a named ascii file
# if the extent is not known (None), the file must be free of comments
# a new numpy array is returned

def load_array_from_ascii_file(file_name, extent = None, data_type = 'real', keyword = None,
                               max_lines_for_keyword = None, comment_char = None,
                               data_free_of_comments = False, skip_c_space = True, use_numbers_only = None):
   """Returns a numpy array with data loaded from an ascii file.

   arguments:
      file_name: string holding name of the existing ascii data file, eg. 'Cell_depth.dat'
      extent: a python list of integers specifying the extent (shape) of the array, eg. [148, 270, 103];
         if None, all data is read and returned as a 'flat' 1D array (data must be free from comments)
      data_type: the type of individual data elements, one of 'real','float','int', 'integer', 'bool' or 'boolean'
      keyword: if present, an attempt is made to find the keyword before reading data, if keyword is None or is
         not found, data is read from the start of the file
      max_lines_for_keyword: can be used to limit the search for keyword (for speed efficiency)
      comment_char: a single character string being the character used to introduce a comment in the file
      data_free_of_comments: if set to True, a faster load is used once any header line comments have been
         skipped
      skip_c_space: if True then a line starting 'C' followed by white space is skipped as a comment
      use_numbers_only: this argument is no longer in use and is ignored

   returns:
      a numpy array of shape specified in extent argument with dtype matching data_type

   example call:
      depth_array = load_data.load_array_from_ascii_file('Cell_depth.dat', [148, 270, 103])

   notes:
      In all use cases, this function is designed to load a single array of data from an ascii file
      that DOES NOT CONTAIN OTHER ARRAYS as well, ie. data for a single simulation keyword in the file.

      If skip_c_space is True, lines starting 'C ' are
      also treated as comments.  If data_free_of_comments is True, there must be
      at least one blank line before the data begins, and no further comments are permitted.
      (This format is designed to handle data files generated by a commonly used geomodelling package.)

      Repeat counts must not be present in the ascii data.

      The extent, if present, can contain any number of dimensions, typically 3 for reservoir modelling work.
      The total number of numbers in the file must match the number of elements in the given extent
      (ie. the product of the list of numbers in the extent argument).
      The order of indices in extent should be 'slowest changing' first, eg.: k,j,i

      The data_type defaults to 'real'
      'real' and 'float' are synonymous; 'int' and 'integer' are synonymous;
      'bool' and 'boolean' are synonymous; default is 'real'
      The numpy data type will be the default 64 bit float or 64 bit int
   """
   # todo: Code enhancement could cater for 32 bit options (and 8 bit for bool) if needed to reduce memory usage

   if extent is None:
      cell_count = -1   # np.fromfile interprets this as 'read everything'
      log.debug('Loading unknown number of array data elements from ascii file ' + file_name)
   else:
      cell_count = pu.cell_count_from_extent(extent)
      log.debug('Loading %1d array data elements from ascii file %s', cell_count, file_name)

   if data_type in ['real', 'float', float]: d_type = 'float'
   elif data_type in ['int', 'integer', int]: d_type = 'int'
   elif data_type in ['bool', 'boolean', bool]: d_type = 'int'  # read booleans as 0 or 1 and convert after
   else:
      assert False, 'Unknown data_type passed to load_array_from_ascii_file' + str(data_type)

   read_file_name = file_name

   with open(read_file_name, 'r') as data_file:

      if not comment_char and not data_free_of_comments:
         comment_char = kf.guess_comment_char(data_file)
         if not comment_char: comment_char = '!'

      if keyword:
         keyword_found = kf.find_keyword(data_file, keyword, max_lines = max_lines_for_keyword)
         if keyword_found: data_file.readline()  # skip keyword line

      kf.skip_blank_lines_and_comments(data_file, comment_char = comment_char, skip_c_space = skip_c_space)

      result = None

      if data_free_of_comments:   # use numpy fromfile function after passing header comments

         result = np.fromfile(data_file, dtype = d_type, count = cell_count, sep = ' ')

         if result is None:
            data_file.seek(0)

      if result is None:

         # builds one very big string, stripping trailing comments

#         s = ''
#         while True:
#            r = data_file.readline()
#            if len(r) == 0: break
#            s += r.partition(comment_char)[0] + ' '
#         result = np.fromstring(s, dtype = d_type, count = cell_count, sep = ' ')
#         # note: extra data will go unnoticed if cell count known!
#         del s

         b = bytearray(data_file.read().encode())
         bc = comment_char.encode()
         nl = b'\n'
         sp = b' '
         i = 0
         while True:
            i = b.find(bc, i)
            if i < 0: break
            eol = b.find(nl, i)
            if eol < 0: eol = len(b)
            b[i : eol] = sp * (eol - i)

         result = np.fromstring(b.decode(), dtype = d_type, count = cell_count, sep = ' ')
         # note: extra data will go unnoticed if cell count known!
         del b

      if result is None:

         assert(extent is not None)   # TODO: remove this restriction by dynamically extending a flat array

         view_1D = np.zeros(extent, dtype = d_type).flatten()
         elements = view_1D.size
         start_of_line = True

         for index in range(elements):

            while True:
               ch = data_file.read(1)
               assert(ch != '')        # premature end of file
               if ch == comment_char:
                  data_file.readline()
                  start_of_line = True
                  continue
               if skip_c_space and start_of_line and ch in ['C', 'c']:
                  next_ch = data_file.read(1)
                  if next_ch in ' \t\n':
                     if next_ch != '\n': data_file.readline()
                     continue
                  ch += next_ch
                  break
               if ch not in ' \t\n':
                  break

            start_of_line = False
            word = ch
            while True:
               ch = data_file.read(1)
               if ch == '':            # end of file (this assumes a whitespace character after last datum)
                  log.error('Not enough data in file %s: %1d of %1d numbers read', file_name, index, elements)
                  assert False, 'not enough data in file'
               if ch in ' \t\n': break
               word += ch
            if ch == '\n': start_of_line = True
            if d_type == 'float': view_1D[index] = float(word)
            else: view_1D[index] = int(word)

         result = view_1D

      kf.skip_blank_lines_and_comments(data_file, comment_char = comment_char, skip_c_space = skip_c_space)
      if not kf.end_of_file(data_file):
         assert False, 'too much data in ascii file: ' + file_name

   if result is not None and extent is not None: result = result.reshape(extent)

   if data_type in ['bool', 'boolean', bool]:
      return result != 0   # convert to boolean
   else:
      return result

# end of load_array_from_ascii_file() def
######################################################################################################


######################################################################################################
# load_fault_mask():
# function to load fault indicator data for cell faces from nexus input deck format ascii file
# indicator array is a numpy array with data type 'bool' and must have been initialised to false
# only faces specified in the input file are modified
# the indicator array is set to true for every face mentioned in the MULT data, regardless of value
# binary data file creation & reuse is supported

def load_fault_mask(file_name, direction, fault_mask, use_binary = False):
   """Modifies a 3D numpy boolean array representing fault locations on cell faces in i,j or k.

   arguments:
      file_name: string giving name of existing ascii file holding nexus MULT keywords in input deck format
      direction: a single character giving direction of interest, must be 'i', 'j' or 'k'
      fault_mask: an existing 3D boolean numpy array which must have been initialised to False everywhere
      use_binary: if True, will attempt to load the boolean mask from a pre-written pure binary file,
                 which will be created if it does not exist

   returns:
      None

   notes :
      see nexus keyword reference for format of MULT keyword

      fault_mask element is set to True if corresponding face appears in the MULT data set,
      whatever function is being applied, and whatever value is specified the function returns None (nothing)
   """

   trans_keyword_dict = {'i':'TX', 'j':'TY', 'k':'TZ'}

   assert(direction in 'ijk')
   assert(fault_mask.ndim == 3)
   mask_extent_kji = fault_mask.shape
   ascii_file_name = file_name
   binary_file_name = file_name

   face_count = pu.cell_count_from_extent(mask_extent_kji)
   log.debug('Loading %1d fault mask elements from file %s', face_count, file_name)

   if use_binary:
      # try reading mask from binary file
      if len(binary_file_name) > 5 and binary_file_name[-5:] == '.' + direction.lower() + '.bb':
         ascii_file_name = ascii_file_name[:-5]
      else:
         binary_file_name = binary_file_name + '.' + direction.lower() + '.bb'
      try:
         with open(binary_file_name, 'rb') as binary_data_file:
            result = np.fromfile(binary_data_file, dtype = 'bool', count = face_count).reshape(mask_extent_kji)
            fault_mask.data = result.data
            # todo: check that end of file has been reached, ie. not too much data in file
            log.info('Fault mask data loaded from binary file %s', binary_file_name)
            return
      except:
         pass

   with open(ascii_file_name, 'r') as data_file:

      while kf.find_keyword_pair(data_file, 'MULT', trans_keyword_dict[direction]):

         line = data_file.readline().upper()
         line = line.partition('!')[0]         # removes trailing comment
         plus_minus_offset = [0,0,0]           # only possibly set to 1 in direction of interest
         if ('MINUS' not in line.split()):     # indicates data apply to 'negative' faces of specified cells
            plus_minus_offset['kji'.index(direction)] = 1

         if not kf.find_keyword(data_file, 'FNAME'): continue  # will mess up if not present for this MULT
         line = data_file.readline().upper()
         words = line.split()

         if not kf.find_number(data_file): continue            # will mess up if no numbers before next MULT

         count = 0
         while kf.number_next(data_file):
            # fetch a new box from the input file (usually the box contains just one cell)
            line = data_file.readline().upper()
            line = line.partition('!')[0]         # removes trailing comment
            words = line.split()
            box_kji0 = box.box_kji0_from_words_iijjkk1(words)
            box_kji0[0] = box_kji0[0] + plus_minus_offset   # adjust min indices
            box_kji0[1] = box_kji0[1] + plus_minus_offset   # adjust max indices
            # check that box is within extent of fault_mask
            for dir in range(2):
               assert(box_kji0[0, dir] >= 0)
               assert(box_kji0[1, dir] < mask_extent_kji[dir])
            # set fault_mask to True for face of every cell in box (usually just one cell, or column)
#            for k in range(box_kji0[0, 0], box_kji0[1, 0] + 1):
#               for j in range(box_kji0[0, 1], box_kji0[1, 1] + 1):
#                  for i in range(box_kji0[0, 2], box_kji0[1, 2] + 1):
#                     if not fault_mask[k,j,i]:
#                        fault_mask[k, j, i] = True
#                        count += 1
            count += (box.volume_of_box(box_kji0)  -
                      np.count_nonzero(fault_mask[box_kji0[0, 0] : box_kji0[1, 0] + 1,
                                                  box_kji0[0, 1] : box_kji0[1, 1] + 1,
                                                  box_kji0[0, 2] : box_kji0[1, 2] + 1]))
            fault_mask[box_kji0[0, 0] : box_kji0[1, 0] + 1,
                       box_kji0[0, 1] : box_kji0[1, 1] + 1,
                       box_kji0[0, 2] : box_kji0[1, 2] + 1] = True


   if use_binary:
      # create binary file for use next time
      try:
         with open(binary_file_name, 'wb') as binary_file_out:
            fault_mask.tofile(binary_file_out)
         log.info('Binary fault mask data file %s created', binary_file_name)
      except:
         log.warn('Failed to write data to binary fault mask file %s', binary_file_name)
         # todo: could delete the binary file in case a corrupt file is left for use next time

# end of load_fault_mask() def
######################################################################################################



import math as maths

import pytest
import numpy as np
from numpy.testing import assert_array_almost_equal

import resqpy.olio.weights_and_measures as wam
from resqpy.olio.exceptions import InvalidUnitError, IncompatibleUnitsError


# ------- Test uoms ------


def test_valid_uoms():

   # Good units
   assert "0.001 gal[US]/gal[US]" in wam.valid_uoms()
   assert "Btu[IT]/min" in wam.valid_uoms()
   assert "%" in wam.valid_uoms()

   # Bad units
   assert "foo barr" not in wam.valid_uoms()
   assert "" not in wam.valid_uoms()
   assert None not in wam.valid_uoms()


def test_uom_aliases():
   for uom, aliases in wam.UOM_ALIASES.items():
      assert uom in wam.valid_uoms(), f"Bad uom {uom}"
      for alias in aliases:
         if alias != uom:
            assert alias not in wam.valid_uoms(), f"Bad alias {alias}"

   for alias, uom in wam.UOM_ALIAS_MAP.items():
      assert uom in wam.valid_uoms(), f"Bad uom {uom}"
      if alias != uom:
         assert alias not in wam.valid_uoms(), f"Bad alias {alias}"

   for uom in wam.CASE_INSENSITIVE_UOMS:
      assert uom in wam.valid_uoms(), f"Bad uom {uom}"


def test_aliases_are_unique():
   n_aliases = sum(map(len, wam.UOM_ALIASES.values()))
   all_aliases = set.union(*wam.UOM_ALIASES.values())
   assert n_aliases == len(all_aliases)


# ------- Test parsing uoms -------


@pytest.mark.parametrize("input_uom, expected_uom", [
   ('gapi', 'gAPI'),
   ('m', 'm'),
   ('M', 'm'),
   ('gal[UK]/mi', 'gal[UK]/mi'),
   ("0.001 gal[US]/gal[US]", "0.001 gal[US]/gal[US]"),
   ('1E6 m3/day', '1E6 m3/d')
])
def test_uom_from_string(input_uom, expected_uom):
   validated_uom = wam.rq_uom(input_uom)
   assert expected_uom == validated_uom


@pytest.mark.parametrize("input_uom", ["foobar", None, "", "qwertyuiol", "bifröst"])
def test_bad_unit_raises_error(input_uom):
   with pytest.raises(InvalidUnitError):
      wam.rq_uom(input_uom)


# ----- Test quantities -------

def test_quantities():
   
   quantities = wam.valid_quantities()
   assert len(quantities) > 10
   for q in quantities:
      uoms = wam.valid_uoms(quantity=q)
      assert len(uoms) > 0
      for uom in uoms:
         assert uom in wam.valid_uoms(), f"Bad uom {uom}"


# ---- Test unit conversions -------

@pytest.mark.parametrize("unit_from, unit_to, value, expected", [
   # Straightforward conversions
   ("m", "m", 1, 1),
   ("m", "km", 1, 0.001),
   ("ft", "m", 1, 0.3048),
   ("ft", "ft[US]", 1, 0.999998),

   # Fractional units with aliases
   ("bbl/d", "stb/day", 1, 1),
   ("mmstb/day", "1E6 bbl/d", 1, 1),
   ("mmscf/day", "1E6 ft3/d", 1, 1),
   ("scf/stb", "ft3/bbl", 1, 1),

   # Aliases of common units
   ("metres", "m", 1, 1),
   ("meters", "m", 1, 1),
   ("pu", "%", 1, 1),
   ("p.u.", "%", 1, 1),

   # Different base units!
   ("%", "v/v", 1, 0.01),
   ("pu", "v/v", 1, 0.01),
   ("m3/m3", "%", 1, 100),
   ("D", "ft2", 10, 1.062315e-10),
])
@pytest.mark.filterwarnings("ignore:Assuming base units")
def test_unit_conversion(unit_from, unit_to, value, expected):
   result = wam.convert(value, unit_from, unit_to)
   assert maths.isclose(result, expected, rel_tol=1e-4)


def test_convert_array():
   # Duck typing should work
   value = np.array([1,2,3])
   expected = np.array([1000, 2000, 3000])
   result = wam.convert(value, unit_from="km", unit_to="m")
   assert_array_almost_equal(result, expected)


def test_conversion_factors_are_numeric():
   for uom in wam.valid_uoms():
      base_unit, dimension, factors = wam.get_conversion_factors(uom)
      assert base_unit in wam.valid_uoms()
      assert len(dimension) > 0
      assert len(factors) == 4, f"Issue with {uom}"
      assert all(isinstance(f, (int, float)) for f in factors), f"Issue with {uom}"


# Test incompatible units raise an Error

@pytest.mark.parametrize("unit_from, unit_to", [
   ("m", "gAPI"),
   ("%", "bbl"),
   ("%", "ft"),
   ("m", "m3"),
])
def test_incompatible_units(unit_from, unit_to):
   with pytest.raises(IncompatibleUnitsError):
      wam.convert(1, unit_from, unit_to)



def test_length_conversion():
   assert maths.isclose(wam.convert_lengths(1.0, 'feet', 'metres'), 0.3048)
   assert maths.isclose(wam.convert_lengths(10.0, 'ft', 'm'), 3.048)
   a = np.random.random((4, 5))
   a_m = 0.3048 * a
   wam.convert_lengths(a, 'ft', 'm')
   assert np.all(np.isclose(a_m, a))
   a = np.random.random((4, 5))
   a_ft = a / 0.3048
   assert np.all(np.isclose(a_ft, wam.convert_lengths(a, 'm', 'ft')))
   b = a.copy()
   wam.convert_lengths(a, 'ft', 'm')
   wam.convert_lengths(a, 'm', 'ft')
   assert np.all(np.isclose(a, b))

def test_time_conversion():
   for period in (1.0, 0.0238, 67.824, 0.0):
      assert maths.isclose(24 * period, wam.convert_times(period, 'days', 'hours'))
      assert maths.isclose(24 * 60 * period, wam.convert_times(period, 'd', 'mins'))
      assert maths.isclose(24 * 60 * 60 * period, wam.convert_times(period, 'day', 'sec'))
      assert maths.isclose(24 * 60 * 60 * 1000 * period, wam.convert_times(period, 'day', 'ms'))
      assert maths.isclose(60 * period, wam.convert_times(period, 'hr', 'min'))
      assert maths.isclose(60 * period, wam.convert_times(period, 'min', 's'))
      assert maths.isclose(period / 24, wam.convert_times(period, 'days', 'hours', invert = True))
      assert maths.isclose(period / (24 * 60), wam.convert_times(period, 'd', 'mins', invert = True))
      assert maths.isclose(period / (24 * 60 * 60), wam.convert_times(period, 'day', 'sec', invert = True))
      assert maths.isclose(period / (24 * 60 * 60 * 1000), wam.convert_times(period, 'day', 'ms', invert = True))
      assert maths.isclose(period / 60, wam.convert_times(period, 'hr', 'min', invert = True))
      assert maths.isclose(period / 60, wam.convert_times(period, 'min', 's', invert = True))
      assert maths.isclose(period / 24, wam.convert_times(period, 'hours', 'days'))
      assert maths.isclose(period / (24 * 60), wam.convert_times(period, 'mins', 'd'))
      assert maths.isclose(period / (24 * 60 * 60), wam.convert_times(period, 'sec', 'day'))
      assert maths.isclose(period / (24 * 60 * 60 * 1000), wam.convert_times(period, 'ms', 'day'))
      assert maths.isclose(period / 60, wam.convert_times(period, 'min', 'hr'))
      assert maths.isclose(period / 60, wam.convert_times(period, 's', 'min'))
   a = np.random.random((2, 3, 4))
   b = a.copy()
   wam.convert_times(a, 'd', 's')
   assert np.all(np.isclose(a, 24.0 * 60.0 * 60.0 * b))
   wam.convert_times(a, 'd', 's', invert = True)
   assert np.all(np.isclose(a, b))

def test_pressure_conversion():
   assert maths.isclose(20684.28, wam.convert_pressures(3000.0, 'psi', 'kPa'), rel_tol = 1.0e-4)
   a = np.random.random(10)
   b = a.copy()
   assert np.all(np.isclose(100.0 * b, wam.convert_pressures(a, 'bar', 'kPa')))
   wam.convert_pressures(a, 'kPa', 'bar')
   assert np.all(np.isclose(b, a))

def test_volume_conversion():
   assert maths.isclose(wam.convert_volumes(1.0, 'm3', 'ft3'), 35.3147, rel_tol = 1.0e-4)
   assert maths.isclose(wam.convert_volumes(1000.0, 'stb', 'm3'), 158.987, rel_tol = 1.0e-4)
   units = ('m3', 'ft3', 'bbl')
   for rate in (2100.0, 0.0, -312.45):
      for from_units in units:
         for to_units in units:
            assert maths.isclose(rate, wam.convert_volumes(wam.convert_volumes(rate, from_units, to_units),
                                                           to_units, from_units))

def test_flow_rate_conversion():
   assert maths.isclose(wam.convert_flow_rates(1000.0, 'm3/d', 'ft3/d'), 35314.6667, rel_tol = 1.0e-5)
   assert maths.isclose(wam.convert_flow_rates(1000.0, 'm3/d', 'bbl/d'), 1000.0 / 0.158987, rel_tol = 1.0e-4)
   units = ('m3/d', 'ft3/s', 'bbl/d', 'stb/hr', '1000 bbl/d', '1E6 ft3/d')
   for rate in (2100.0, 0.0, -312.45):
      for from_units in units:
         for to_units in units:
            assert maths.isclose(rate, wam.convert_flow_rates(wam.convert_flow_rates(rate, from_units, to_units),
                                                              to_units, from_units))
